"""Google Gemini Interactions HTTP provider.

Uses direct synchronous HTTP calls to the Gemini Interactions API.
Conversation history is managed server-side via ``previous_interaction_id``.
"""

from __future__ import annotations

import json
import time
import urllib.error
import urllib.request
from typing import Any

from pbi_agent import __version__
from pbi_agent.agent.system_prompt import get_system_prompt
from pbi_agent.agent.tool_runtime import execute_tool_calls as _execute_tool_calls
from pbi_agent.config import Settings, missing_api_key_message
from pbi_agent.models.messages import (
    CompletedResponse,
    ImageAttachment,
    TokenUsage,
    ToolCall,
    UserTurnInput,
    WebSearchSource,
)
from pbi_agent.providers.base import Provider
from pbi_agent.tools.registry import get_openai_tool_definitions
from pbi_agent.tools.types import ToolContext
from pbi_agent.ui.display_protocol import DisplayProtocol

_REQUEST_TIMEOUT_SECS = 3600.0
_THINKING_LEVEL_MAP: dict[str, str] = {
    "low": "low",
    "medium": "medium",
    "high": "high",
    "xhigh": "high",
}

_HTTP_ERROR_TYPES: dict[int, str] = {
    400: "invalid_argument",
    403: "permission_denied",
    404: "not_found",
    429: "resource_exhausted",
    500: "internal",
    503: "unavailable",
    504: "deadline_exceeded",
}

_HTTP_ERROR_MESSAGES: dict[int, str] = {
    400: "The request body is malformed.",
    403: "Your API key doesn't have the required permissions.",
    404: "The requested resource wasn't found.",
    429: "You've exceeded the rate limit.",
    500: "An unexpected error occurred on Google's side.",
    503: "The service may be temporarily overloaded or down.",
    504: "The service is unable to finish processing within the deadline.",
}


class GoogleProvider(Provider):
    """Provider backed by the Gemini Interactions HTTP API."""

    def __init__(
        self,
        settings: Settings,
        *,
        system_prompt: str | None = None,
        excluded_tools: set[str] | None = None,
    ) -> None:
        self._settings = settings
        self._tools = _google_tool_definitions(excluded_names=excluded_tools)
        if settings.web_search:
            self._tools.append({"type": "google_search"})
        self._instructions = system_prompt or get_system_prompt()
        self._previous_interaction_id: str | None = None

    @property
    def settings(self) -> Settings:
        return self._settings

    def set_previous_response_id(self, response_id: str | None) -> None:
        self._previous_interaction_id = response_id

    def connect(self) -> None:
        if not self._settings.api_key:
            raise ValueError(missing_api_key_message("google"))

    def close(self) -> None:
        pass

    def reset_conversation(self) -> None:
        self._previous_interaction_id = None

    def request_turn(
        self,
        *,
        user_message: str | None = None,
        user_input: UserTurnInput | None = None,
        tool_result_items: list[dict[str, Any]] | None = None,
        instructions: str | None = None,
        display: DisplayProtocol,
        session_usage: TokenUsage,
        turn_usage: TokenUsage,
    ) -> CompletedResponse:
        if user_input is None and user_message is not None:
            user_input = UserTurnInput(text=user_message)

        if user_input is not None:
            input_value = _google_user_input_value(user_input)
        elif tool_result_items is not None:
            input_value = tool_result_items
        else:
            raise ValueError("Either user_input or tool_result_items is required")

        result = self._http_request(
            input_value=input_value,
            instructions=instructions or self._instructions,
            display=display,
        )
        self._previous_interaction_id = result.response_id
        session_usage.add(result.usage)
        turn_usage.add(result.usage)
        display.session_usage(session_usage)

        if result.reasoning_summary or result.reasoning_content:
            display.render_thinking(
                _reasoning_body_text(
                    result.reasoning_content,
                    result.reasoning_summary,
                ),
                title=result.reasoning_summary or None,
            )

        display_items = result.provider_data.get("display_items")
        if isinstance(display_items, list) and display_items:
            for item in display_items:
                if not isinstance(item, dict):
                    continue
                item_type = item.get("type")
                if item_type == "text":
                    text = item.get("text")
                    if isinstance(text, str) and text:
                        display.render_markdown(text)
                elif item_type == "google_search_result":
                    _display_web_search_result(
                        display,
                        item.get("sources", []),
                        queries=item.get("queries", []),
                    )
        else:
            if result.web_search_sources:
                _display_web_search_result(
                    display,
                    result.web_search_sources,
                    queries=result.provider_data.get("web_search_queries", []),
                )
            if result.text:
                display.render_markdown(result.text)

        return result

    def execute_tool_calls(
        self,
        response: CompletedResponse,
        *,
        max_workers: int,
        display: DisplayProtocol,
        session_usage: TokenUsage,
        turn_usage: TokenUsage,
        sub_agent_depth: int = 0,
    ) -> tuple[list[dict[str, Any]], bool]:
        if not response.function_calls:
            return [], False

        displayable_calls = [
            call for call in response.function_calls if call.name != "sub_agent"
        ]
        if displayable_calls:
            display.function_start(len(displayable_calls))
        batch = _execute_tool_calls(
            response.function_calls,
            max_workers=max_workers,
            context=ToolContext(
                settings=self._settings,
                display=display,
                session_usage=session_usage,
                turn_usage=turn_usage,
                sub_agent_depth=sub_agent_depth,
            ),
        )

        tool_result_items: list[dict[str, Any]] = []
        for result in batch.results:
            call = _find_by_id(response.function_calls, result.call_id)
            if not (call and call.name == "sub_agent"):
                display.function_result(
                    name=call.name if call else "unknown",
                    success=not result.is_error,
                    call_id=result.call_id,
                    arguments=call.arguments if call else None,
                )
            item: dict[str, Any] = {
                "type": "function_result",
                "name": call.name if call else "",
                "call_id": result.call_id,
                "result": _google_function_result_value(result),
            }
            if result.is_error:
                item["is_error"] = True
            tool_result_items.append(item)
        if displayable_calls:
            display.tool_group_end()

        return tool_result_items, batch.had_errors

    def _http_request(
        self,
        *,
        input_value: str | list[dict[str, Any]],
        instructions: str,
        display: DisplayProtocol,
    ) -> CompletedResponse:
        display.wait_start(_waiting_message_for_input(input_value))

        body = self._build_request_body(
            input_value=input_value,
            instructions=instructions,
        )
        request_data = json.dumps(body).encode("utf-8")
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "x-goog-api-key": self._settings.api_key,
            "User-Agent": f"pbi-agent/{__version__}",
        }

        max_retries = self._settings.max_retries
        last_error: Exception | None = None
        last_error_message: str | None = None

        for attempt in range(max_retries + 1):
            if attempt > 0:
                display.retry_notice(attempt, max_retries)

            try:
                req = urllib.request.Request(
                    self._settings.responses_url,
                    data=request_data,
                    headers=headers,
                    method="POST",
                )
                with urllib.request.urlopen(req, timeout=_REQUEST_TIMEOUT_SECS) as resp:
                    response_json = json.loads(resp.read().decode("utf-8"))

                _raise_if_interaction_failed(response_json)
                result = self._parse_response(response_json)
                display.wait_stop()
                return result
            except urllib.error.HTTPError as exc:
                error_body = _read_error_body(exc)
                error_payload = _normalize_http_error(exc, error_body)
                if exc.code == 429:
                    if attempt >= max_retries:
                        display.wait_stop()
                        raise RuntimeError(
                            _format_error_message(
                                f"Google rate limit exceeded after {max_retries + 1} attempts",
                                error_payload,
                            )
                        ) from exc
                    wait = _extract_retry_after(exc, attempt)
                    display.rate_limit_notice(
                        wait_seconds=wait,
                        attempt=attempt + 1,
                        max_retries=max_retries,
                    )
                    time.sleep(wait)
                    continue

                if exc.code == 503:
                    if attempt >= max_retries:
                        display.wait_stop()
                        raise RuntimeError(
                            _format_error_message(
                                f"Google API overloaded after {max_retries + 1} attempts",
                                error_payload,
                            )
                        ) from exc
                    wait = min(2.0 * (2**attempt), 30.0) + 1.0
                    display.overload_notice(
                        wait_seconds=wait,
                        attempt=attempt + 1,
                        max_retries=max_retries,
                    )
                    time.sleep(wait)
                    continue

                if exc.code >= 500:
                    last_error = exc
                    last_error_message = _format_error_message(
                        f"Google request failed after {max_retries + 1} attempts",
                        error_payload,
                    )
                    continue

                display.wait_stop()
                raise RuntimeError(
                    _format_error_message(
                        f"Google Interactions API error {exc.code}",
                        error_payload,
                    )
                ) from exc
            except urllib.error.URLError as exc:
                last_error = exc
                last_error_message = None
                continue

        display.wait_stop()
        if last_error is not None:
            if last_error_message:
                raise RuntimeError(last_error_message) from last_error
            raise RuntimeError(
                f"Google request failed after {max_retries + 1} attempts: {last_error}"
            ) from last_error
        raise RuntimeError("Google request failed after retries.")

    def _build_request_body(
        self,
        *,
        input_value: str | list[dict[str, Any]],
        instructions: str | None,
    ) -> dict[str, Any]:
        body: dict[str, Any] = {
            "model": self._settings.model,
            "input": input_value,
            "tools": self._tools,
            "stream": False,
            "store": True,
            "generation_config": {
                "thinking_level": _THINKING_LEVEL_MAP.get(
                    self._settings.reasoning_effort,
                    "high",
                ),
                "thinking_summaries": "auto",
                "max_output_tokens": self._settings.max_tokens,
            },
        }
        if instructions:
            body["system_instruction"] = instructions
        if self._previous_interaction_id:
            body["previous_interaction_id"] = self._previous_interaction_id
        return body

    def _parse_response(self, response_json: dict[str, Any]) -> CompletedResponse:
        text_parts: list[str] = []
        thought_summary_parts: list[str] = []
        thought_signatures: list[str] = []
        function_calls: list[ToolCall] = []
        display_items: list[dict[str, Any]] = []
        saw_google_search_result = False
        pending_google_search_queries: dict[str, list[str]] = {}

        output_items = response_json.get("outputs", [])
        if not isinstance(output_items, list):
            output_items = []

        for item in output_items:
            if not isinstance(item, dict):
                continue
            item_type = item.get("type")

            if item_type == "text":
                text = item.get("text", "")
                if isinstance(text, str) and text:
                    text_parts.append(text)
                    display_items.append({"type": "text", "text": text})

            elif item_type == "thought":
                summary_text = _extract_thought_summary_text(item.get("summary"))
                if summary_text:
                    thought_summary_parts.append(summary_text)
                signature = item.get("signature")
                if isinstance(signature, str) and signature:
                    thought_signatures.append(signature)

            elif item_type == "function_call":
                function_calls.append(
                    ToolCall(
                        call_id=str(item.get("id", "")),
                        name=str(item.get("name", "")),
                        arguments=item.get("arguments"),
                    )
                )
            elif item_type == "google_search_call":
                call_id = str(item.get("id", ""))
                if call_id:
                    pending_google_search_queries[call_id] = (
                        _extract_google_search_call_queries(item)
                    )
            elif item_type == "google_search_result":
                saw_google_search_result = True
                display_items.append(
                    {
                        "type": "google_search_result",
                        "queries": pending_google_search_queries.get(
                            str(item.get("call_id", "")),
                            _extract_google_search_queries(item, response_json),
                        ),
                        "sources": _extract_google_search_result_sources(
                            item,
                            response_json,
                        ),
                    }
                )

        web_search_queries = _extract_google_search_queries(None, response_json)
        web_search_sources = _extract_google_search_sources(response_json)
        if web_search_sources and not saw_google_search_result:
            insert_at = next(
                (
                    index
                    for index, item in enumerate(display_items)
                    if isinstance(item, dict) and item.get("type") == "text"
                ),
                len(display_items),
            )
            display_items.insert(
                insert_at,
                {
                    "type": "google_search_result",
                    "queries": web_search_queries,
                    "sources": web_search_sources,
                },
            )

        usage_obj = response_json.get("usage", {})
        input_tokens = int(_usage_value(usage_obj, "total_input_tokens"))
        cached_input_tokens = int(_usage_value(usage_obj, "total_cached_tokens"))
        output_tokens = int(_usage_value(usage_obj, "total_output_tokens"))
        reasoning_tokens = int(_usage_value(usage_obj, "total_thought_tokens"))
        tool_use_tokens = int(_usage_value(usage_obj, "total_tool_use_tokens"))
        provider_total_tokens = int(_usage_value(usage_obj, "total_tokens"))

        return CompletedResponse(
            response_id=response_json.get("id"),
            text="\n\n".join(part for part in text_parts if part.strip()).strip(),
            usage=TokenUsage(
                input_tokens=input_tokens,
                cached_input_tokens=cached_input_tokens,
                output_tokens=output_tokens,
                reasoning_tokens=reasoning_tokens,
                tool_use_tokens=tool_use_tokens,
                provider_total_tokens=provider_total_tokens,
                context_tokens=provider_total_tokens,
                model=_response_model_name(response_json),
            ),
            function_calls=function_calls,
            reasoning_content="\n\n".join(
                part for part in thought_summary_parts if part.strip()
            ).strip(),
            provider_data={
                "status": response_json.get("status"),
                "thought_signatures": thought_signatures,
                "display_items": display_items,
                "web_search_queries": web_search_queries,
            },
            web_search_sources=web_search_sources,
        )


def _extract_google_search_sources(
    response_json: dict[str, Any],
) -> list[WebSearchSource]:
    """Extract web search sources from Google Interactions and grounding metadata."""
    sources: list[WebSearchSource] = []

    output_items = response_json.get("outputs", [])
    if isinstance(output_items, list):
        for item in output_items:
            if not isinstance(item, dict) or item.get("type") != "google_search_result":
                continue
            sources.extend(_extract_google_search_result_sources(item, response_json))

    if sources:
        return sources

    sources.extend(_extract_google_text_annotation_sources(response_json))
    if sources:
        return sources

    metadata = _extract_google_grounding_metadata(response_json)
    if not metadata:
        return []
    chunks = metadata.get("groundingChunks", [])
    if not isinstance(chunks, list):
        return []
    for chunk in chunks:
        if not isinstance(chunk, dict):
            continue
        web = chunk.get("web")
        if not isinstance(web, dict):
            continue
        url = str(web.get("uri", ""))
        if not url:
            continue
        sources.append(
            WebSearchSource(
                title=str(web.get("title", "")),
                url=url,
            )
        )
    return sources


def _extract_google_search_result_sources(
    item: dict[str, Any],
    response_json: dict[str, Any],
) -> list[WebSearchSource]:
    results = item.get("result", [])
    sources: list[WebSearchSource] = []
    if isinstance(results, list):
        for entry in results:
            if not isinstance(entry, dict):
                continue
            url = str(entry.get("url", ""))
            if not url:
                continue
            sources.append(
                WebSearchSource(
                    title=str(entry.get("title", "")),
                    url=url,
                )
            )
    if sources:
        return sources
    return _extract_google_text_annotation_sources(response_json)


def _extract_google_text_annotation_sources(
    response_json: dict[str, Any],
) -> list[WebSearchSource]:
    output_items = response_json.get("outputs", [])
    if not isinstance(output_items, list):
        return []
    sources: list[WebSearchSource] = []
    for item in output_items:
        if not isinstance(item, dict) or item.get("type") != "text":
            continue
        annotations = item.get("annotations", [])
        if not isinstance(annotations, list):
            continue
        for annotation in annotations:
            if not isinstance(annotation, dict):
                continue
            if annotation.get("type") != "url_citation":
                continue
            url = str(annotation.get("url", ""))
            if not url:
                continue
            sources.append(
                WebSearchSource(
                    title=str(annotation.get("title", "")),
                    url=url,
                )
            )
    return sources
    return sources


def _extract_google_grounding_metadata(
    response_json: dict[str, Any],
) -> dict[str, Any] | None:
    metadata = response_json.get("groundingMetadata")
    if isinstance(metadata, dict):
        return metadata

    candidates = response_json.get("candidates")
    if isinstance(candidates, list):
        for candidate in candidates:
            if not isinstance(candidate, dict):
                continue
            metadata = candidate.get("groundingMetadata")
            if isinstance(metadata, dict):
                return metadata

    output_items = response_json.get("outputs", [])
    if isinstance(output_items, list):
        for item in output_items:
            if not isinstance(item, dict):
                continue
            metadata = item.get("groundingMetadata")
            if isinstance(metadata, dict):
                return metadata

    return None


def _extract_google_search_queries(
    item: dict[str, Any] | None,
    response_json: dict[str, Any],
) -> list[str]:
    if isinstance(item, dict):
        raw_queries = item.get("web_search_queries") or item.get("webSearchQueries")
        if isinstance(raw_queries, list):
            queries = [
                str(query).strip() for query in raw_queries if str(query).strip()
            ]
            if queries:
                return queries

    metadata = _extract_google_grounding_metadata(response_json)
    if not metadata:
        return []
    raw_queries = metadata.get("webSearchQueries", [])
    if not isinstance(raw_queries, list):
        return []
    return [str(query).strip() for query in raw_queries if str(query).strip()]


def _extract_google_search_call_queries(item: dict[str, Any]) -> list[str]:
    arguments = item.get("arguments")
    if not isinstance(arguments, dict):
        return []
    raw_queries = arguments.get("queries", [])
    if not isinstance(raw_queries, list):
        return []
    return [str(query).strip() for query in raw_queries if str(query).strip()]


def _display_web_search_result(
    display: DisplayProtocol,
    sources: list[WebSearchSource] | list[dict[str, Any]],
    *,
    queries: list[str] | None = None,
) -> None:
    normalized_sources = [
        source
        if isinstance(source, dict)
        else {"title": source.title, "url": source.url, "snippet": source.snippet}
        for source in (sources or [])
    ]
    display.function_start(1)
    display.function_result(
        name="web_search",
        success=True,
        call_id="",
        arguments={
            "sources": normalized_sources,
            "queries": list(queries or []),
        },
    )
    display.tool_group_end()


def _find_by_id(calls: list[ToolCall], call_id: str) -> ToolCall | None:
    for call in calls:
        if call.call_id == call_id:
            return call
    return None


def _google_user_input_value(
    user_input: UserTurnInput,
) -> str | list[dict[str, Any]]:
    if not user_input.images:
        return user_input.text

    content: list[dict[str, Any]] = []
    if user_input.text:
        content.append({"type": "text", "text": user_input.text})
    for image in user_input.images:
        content.append(_google_image_part(image))
    return content


def _google_function_result_value(result) -> str | list[dict[str, Any]]:
    if not result.attachments:
        return result.output_json

    content = [{"type": "text", "text": result.output_json}]
    content.extend(_google_image_part(image) for image in result.attachments)
    return content


def _google_image_part(image: ImageAttachment) -> dict[str, Any]:
    return {
        "type": "image",
        "mime_type": image.mime_type,
        "data": image.data_base64,
    }


def _google_tool_definitions(
    *, excluded_names: set[str] | None = None
) -> list[dict[str, Any]]:
    return [
        _normalize_google_tool_definition(tool)
        for tool in get_openai_tool_definitions(excluded_names=excluded_names)
    ]


def _normalize_google_tool_definition(tool: dict[str, Any]) -> dict[str, Any]:
    normalized = dict(tool)
    parameters = normalized.get("parameters")
    if isinstance(parameters, dict):
        normalized["parameters"] = _normalize_google_schema(parameters)
    return normalized


def _normalize_google_schema(value: Any) -> Any:
    if isinstance(value, dict):
        normalized: dict[str, Any] = {}
        for key, child in value.items():
            if key == "required" and isinstance(child, list) and not child:
                continue
            normalized[key] = _normalize_google_schema(child)
        return normalized
    if isinstance(value, list):
        return [_normalize_google_schema(item) for item in value]
    return value


def _raise_if_interaction_failed(response_json: dict[str, Any]) -> None:
    status = response_json.get("status")
    if status in {"failed", "cancelled", "incomplete"}:
        error_obj = response_json.get("error")
        if isinstance(error_obj, dict):
            code = str(error_obj.get("code", "unknown_error"))
            message = str(error_obj.get("message", "No error message"))
            raise RuntimeError(f"Google interaction failed ({code}): {message}")
        raise RuntimeError(f"Google interaction failed with status {status}.")
    if status == "in_progress":
        raise RuntimeError(
            "Google interaction returned in_progress for a non-background request."
        )


def _extract_thought_summary_text(summary: Any) -> str:
    if isinstance(summary, str):
        return summary.strip()
    if isinstance(summary, dict):
        return _extract_text_content(summary).strip()
    if not isinstance(summary, list):
        return ""

    parts: list[str] = []
    for item in summary:
        if isinstance(item, str):
            if item.strip():
                parts.append(item.strip())
            continue
        if isinstance(item, dict):
            text = _extract_text_content(item).strip()
            if text:
                parts.append(text)
    return "\n\n".join(parts).strip()


def _extract_text_content(item: dict[str, Any]) -> str:
    if item.get("type") == "text":
        text = item.get("text", "")
        return text if isinstance(text, str) else ""
    content = item.get("content")
    if isinstance(content, list):
        parts: list[str] = []
        for child in content:
            if not isinstance(child, dict):
                continue
            text = _extract_text_content(child)
            if text:
                parts.append(text)
        return "\n\n".join(parts)
    return ""


def _usage_value(usage_obj: Any, key: str) -> int:
    if not isinstance(usage_obj, dict):
        return 0
    return int(usage_obj.get(key, 0) or 0)


def _response_model_name(response_json: dict[str, Any]) -> str:
    model = response_json.get("model")
    return model if isinstance(model, str) else ""


def _read_error_body(exc: urllib.error.HTTPError) -> str:
    try:
        return exc.read().decode("utf-8", errors="replace")
    except Exception:
        return ""


def _normalize_http_error(
    exc: urllib.error.HTTPError,
    error_body: str,
) -> dict[str, Any]:
    payload = _parse_error_payload(error_body)
    error_type = _HTTP_ERROR_TYPES.get(exc.code)
    message = _HTTP_ERROR_MESSAGES.get(exc.code, f"HTTP {exc.code}")
    request_id = _request_id_from_headers(exc)

    if payload is not None:
        error_value = payload.get("error")
        if isinstance(error_value, dict):
            payload_type = error_value.get("status")
            if isinstance(payload_type, str) and payload_type.strip():
                error_type = payload_type.strip().lower()
            payload_message = error_value.get("message")
            if isinstance(payload_message, str) and payload_message.strip():
                message = payload_message.strip()
            payload_request_id = error_value.get("request_id")
            if isinstance(payload_request_id, str) and payload_request_id.strip():
                request_id = payload_request_id.strip()

    if error_type is None:
        if 400 <= exc.code < 500:
            error_type = "invalid_argument"
        else:
            error_type = "internal"

    normalized: dict[str, Any] = {
        "type": "error",
        "status": exc.code,
        "error": {
            "type": error_type,
            "message": message,
        },
    }
    if request_id:
        normalized["request_id"] = request_id
    return normalized


def _parse_error_payload(error_body: str) -> dict[str, Any] | None:
    stripped = error_body.strip()
    if not stripped.startswith("{"):
        return None
    try:
        payload = json.loads(stripped)
    except json.JSONDecodeError:
        return None
    return payload if isinstance(payload, dict) else None


def _request_id_from_headers(exc: urllib.error.HTTPError) -> str | None:
    if not exc.headers:
        return None
    for header in ("x-request-id", "request-id", "x-goog-request-id"):
        request_id = exc.headers.get(header)
        if isinstance(request_id, str) and request_id.strip():
            return request_id.strip()
    return None


def _format_error_message(prefix: str, error_payload: dict[str, Any]) -> str:
    return f"{prefix}: {json.dumps(error_payload, sort_keys=True)}"


def _extract_retry_after(exc: urllib.error.HTTPError, attempt: int) -> float:
    try:
        retry_after = exc.headers.get("Retry-After") if exc.headers else None
        if retry_after:
            return max(0.1, min(float(retry_after), 60.0)) + 1.0
    except (TypeError, ValueError):
        pass
    return min(2.0 * (2**attempt), 30.0) + 1.0


def _waiting_message_for_input(input_value: str | list[dict[str, Any]]) -> str:
    if isinstance(input_value, str):
        return "analyzing your request..."

    has_tool_output = any(
        isinstance(item, dict) and item.get("type") == "function_result"
        for item in input_value
    )
    if has_tool_output:
        return "integrating tool results..."

    return "processing..."


def _reasoning_body_text(reasoning_text: str, summary_text: str) -> str | None:
    if reasoning_text.strip():
        return reasoning_text
    body = _summary_body_text(summary_text)
    return body if body.strip() else None


def _summary_body_text(summary_text: str) -> str:
    lines = summary_text.splitlines()
    first_non_empty_index = next(
        (idx for idx, line in enumerate(lines) if line.strip()),
        None,
    )
    if first_non_empty_index is None:
        return ""
    remaining_non_empty = any(
        line.strip() for line in lines[first_non_empty_index + 1 :]
    )
    if not remaining_non_empty:
        return ""
    return "\n".join(lines[first_non_empty_index + 1 :]).lstrip("\r\n")



from __future__ import annotations

import os
import re
from pathlib import Path
from typing import Any, Callable, Iterator

from pbi_agent.tools.output import bound_output
from pbi_agent.tools.types import ToolContext, ToolSpec
from pbi_agent.tools.workspace_filters import build_glob_matcher
from pbi_agent.tools.workspace_filters import should_skip_directory_name
from pbi_agent.tools.workspace_access import DEFAULT_MAX_MATCHES
from pbi_agent.tools.workspace_access import normalize_positive_int
from pbi_agent.tools.workspace_access import open_text_file
from pbi_agent.tools.workspace_access import relative_workspace_path
from pbi_agent.tools.workspace_access import resolve_safe_path

SPEC = ToolSpec(
    name="search_files",
    description=(
        "Search text files in the workspace for a string or regex pattern. "
        "Use this for safe cross-platform content search instead of shell grep."
    ),
    parameters_schema={
        "type": "object",
        "properties": {
            "pattern": {
                "type": "string",
                "description": "String or regex pattern to search for.",
            },
            "path": {
                "type": "string",
                "description": (
                    "Directory or file path relative to the workspace root "
                    "(or absolute within workspace). Defaults to '.'."
                ),
            },
            "glob": {
                "type": "string",
                "description": (
                    "Optional glob filter for candidate files. Match against the "
                    "entry name unless the pattern includes a path separator."
                ),
            },
            "regex": {
                "type": "boolean",
                "description": "Treat pattern as a regular expression.",
                "default": False,
            },
            "max_matches": {
                "type": "integer",
                "description": "Maximum number of matches to return. Defaults to 100.",
            },
        },
        "required": ["pattern"],
        "additionalProperties": False,
    },
)


def handle(arguments: dict[str, Any], context: ToolContext) -> dict[str, Any]:
    del context
    pattern = arguments.get("pattern", "")
    if not isinstance(pattern, str) or not pattern:
        return {"error": "'pattern' must be a non-empty string."}

    root = Path.cwd().resolve()
    regex_enabled = bool(arguments.get("regex", False))
    max_matches = normalize_positive_int(
        arguments.get("max_matches"),
        default=DEFAULT_MAX_MATCHES,
    )
    glob_pattern = arguments.get("glob")

    try:
        target_path = resolve_safe_path(root, arguments.get("path"))
        matcher = _build_matcher(pattern, regex_enabled)
        glob_matcher = build_glob_matcher(glob_pattern)
        matches: list[dict[str, Any]] = []

        for candidate, relative_path in _iter_candidate_files(
            root,
            target_path,
            glob_matcher,
        ):
            try:
                with open_text_file(candidate) as text_handle:
                    for line_number, line in enumerate(text_handle, start=1):
                        line_text = line.rstrip("\r\n")
                        if not matcher(line_text):
                            continue
                        match = {
                            "path": relative_path,
                            "line_number": line_number,
                            "line": line_text,
                        }
                        bounded_match = _bound_match_fields(match)
                        matches.append(bounded_match)
                        if len(matches) >= max_matches:
                            return {
                                "matches": matches,
                                "matches_truncated": True,
                            }
            except ValueError as exc:
                if str(exc).startswith("binary file is not supported:"):
                    continue
                raise

        return {
            "matches": matches,
        }
    except Exception as exc:
        return {"error": bound_output(str(exc))[0]}


def _build_matcher(pattern: str, regex_enabled: bool) -> Any:
    if not regex_enabled:
        return lambda line: pattern in line

    compiled = re.compile(pattern)
    return lambda line: compiled.search(line) is not None


def _iter_candidate_files(
    root: Path,
    target_path: Path,
    glob_matcher: Callable[[str, str], bool],
) -> Iterator[tuple[Path, str]]:
    if not target_path.exists():
        raise FileNotFoundError(f"path not found: {target_path}")

    if target_path.is_file():
        relative_path = relative_workspace_path(root, target_path)
        if glob_matcher(relative_path, target_path.name):
            yield target_path, relative_path
        return

    if not target_path.is_dir():
        raise ValueError(f"path is not a regular file or directory: {target_path}")

    for current_root, dirnames, filenames in os.walk(
        target_path,
        topdown=True,
        followlinks=False,
    ):
        dirnames[:] = [
            dirname
            for dirname in sorted(dirnames, key=str.casefold)
            if not should_skip_directory_name(dirname)
        ]
        filenames.sort(key=str.casefold)

        current = Path(current_root)
        for filename in filenames:
            candidate = current / filename
            relative_path = candidate.relative_to(root).as_posix()
            if not glob_matcher(relative_path, filename):
                continue
            yield candidate, relative_path


def _bound_match_fields(match: dict[str, Any]) -> dict[str, Any]:
    bounded_path, path_truncated = bound_output(str(match["path"]))
    bounded_line, line_truncated = bound_output(str(match["line"]))
    payload = {
        **match,
        "path": bounded_path,
        "line": bounded_line,
    }
    if path_truncated:
        payload["path_truncated"] = True
    if line_truncated:
        payload["line_truncated"] = True
    return payload



from __future__ import annotations

from contextlib import contextmanager
from pathlib import Path

from pbi_agent.tools import search_files as search_files_tool
from pbi_agent.tools.output import MAX_OUTPUT_CHARS
from pbi_agent.tools.types import ToolContext


def test_search_files_finds_matches_and_skips_binary_files(
    tmp_path: Path,
    monkeypatch,
) -> None:
    monkeypatch.chdir(tmp_path)
    (tmp_path / "notes").mkdir()
    (tmp_path / "notes" / "one.txt").write_text(
        "alpha\nneedle here\n", encoding="utf-8"
    )
    (tmp_path / "notes" / "two.md").write_text("needle again\n", encoding="utf-8")
    (tmp_path / "notes" / "blob.bin").write_bytes(b"\x00\x01\x02needle")

    result = search_files_tool.handle(
        {"pattern": "needle", "path": "notes"},
        ToolContext(),
    )

    assert result["matches"] == [
        {
            "path": "notes/one.txt",
            "line_number": 2,
            "line": "needle here",
        },
        {
            "path": "notes/two.md",
            "line_number": 1,
            "line": "needle again",
        },
    ]


def test_search_files_supports_regex_and_bounds_long_match_lines(
    tmp_path: Path,
    monkeypatch,
) -> None:
    monkeypatch.chdir(tmp_path)
    long_line = f"prefix-{'x' * (MAX_OUTPUT_CHARS + 200)}-suffix"
    (tmp_path / "report.txt").write_text(f"{long_line}\nitem-42\n", encoding="utf-8")

    result = search_files_tool.handle(
        {"pattern": r"item-\d+|prefix-", "regex": True, "max_matches": 1},
        ToolContext(),
    )

    assert result["matches_truncated"] is True
    match = result["matches"][0]
    assert match["path"] == "report.txt"
    assert match["line_truncated"] is True
    assert len(match["line"]) <= MAX_OUTPUT_CHARS
    assert match["line"].startswith("prefix-")
    assert match["line"].endswith("-suffix")
    assert "chars omitted" in match["line"]


def test_search_files_rejects_paths_outside_workspace(
    tmp_path: Path,
    monkeypatch,
) -> None:
    monkeypatch.chdir(tmp_path)
    outside_path = tmp_path.parent

    result = search_files_tool.handle(
        {"pattern": "needle", "path": str(outside_path)},
        ToolContext(),
    )

    assert "outside workspace" in result["error"]


def test_search_files_stops_after_max_matches_without_exhausting_candidates(
    tmp_path: Path,
    monkeypatch,
) -> None:
    monkeypatch.chdir(tmp_path)
    first = tmp_path / "first.txt"
    second = tmp_path / "second.txt"
    first.write_text("needle here\n", encoding="utf-8")
    second.write_text("needle there\n", encoding="utf-8")

    def fake_iter_candidate_files(root: Path, target_path: Path, glob_matcher):
        del root, target_path, glob_matcher
        yield first, "first.txt"
        yield second, "second.txt"

    @contextmanager
    def fake_open_text_file(path: Path, *, encoding: str = "auto"):
        del encoding
        if path == second:
            raise AssertionError("search_files should stop before opening second.txt")
        with path.open("r", encoding="utf-8") as handle:
            yield handle

    monkeypatch.setattr(
        search_files_tool, "_iter_candidate_files", fake_iter_candidate_files
    )
    monkeypatch.setattr(search_files_tool, "open_text_file", fake_open_text_file)

    result = search_files_tool.handle(
        {"pattern": "needle", "max_matches": 1},
        ToolContext(),
    )

    assert result["matches_truncated"] is True
    assert result["matches"] == [
        {
            "path": "first.txt",
            "line_number": 1,
            "line": "needle here",
        }
    ]


def test_search_files_does_not_echo_input_parameters(
    tmp_path: Path,
    monkeypatch,
) -> None:
    monkeypatch.chdir(tmp_path)
    (tmp_path / "notes").mkdir()
    (tmp_path / "notes" / "one.txt").write_text("needle here\n", encoding="utf-8")

    result = search_files_tool.handle(
        {
            "pattern": "needle",
            "path": "notes",
            "glob": "*.txt",
            "regex": False,
            "max_matches": 20,
        },
        ToolContext(),
    )

    assert "pattern" not in result
    assert "path" not in result
    assert "glob" not in result
    assert "regex" not in result
    assert "searched_files" not in result
    assert "skipped_binary_files" not in result


def test_search_files_skips_filtered_directories_during_recursive_search(
    tmp_path: Path,
    monkeypatch,
) -> None:
    monkeypatch.chdir(tmp_path)
    (tmp_path / "src").mkdir()
    (tmp_path / "src" / "app.py").write_text("needle in src\n", encoding="utf-8")
    (tmp_path / ".venv").mkdir()
    (tmp_path / ".venv" / "ignored.py").write_text(
        "needle in venv\n",
        encoding="utf-8",
    )

    result = search_files_tool.handle(
        {"pattern": "needle", "path": ".", "glob": "*.py", "max_matches": 10},
        ToolContext(),
    )

    assert result["matches"] == [
        {
            "path": "src/app.py",
            "line_number": 1,
            "line": "needle in src",
        }
    ]



from __future__ import annotations

import csv
import io
import math
from zipfile import BadZipFile
from datetime import date, datetime, time
from pathlib import Path
from typing import Any

from pbi_agent.tools.output import bound_output
from pbi_agent.tools.types import ToolContext, ToolSpec
from pbi_agent.tools.workspace_access import DEFAULT_MAX_LINES
from pbi_agent.tools.workspace_access import normalize_positive_int
from pbi_agent.tools.workspace_access import open_text_file
from pbi_agent.tools.workspace_access import relative_workspace_path
from pbi_agent.tools.workspace_access import resolve_safe_path

MAX_READ_FILE_OUTPUT_CHARS = 12_000
DATAFRAME_PREVIEW_ROWS = 3
MAX_TABULAR_SCHEMA_CHARS = 8_000
MAX_TABULAR_PREVIEW_CHARS = 2_500

_TABULAR_EXTENSIONS = {
    ".csv",
    ".tsv",
    ".xlsx",
    ".xls",
    ".parquet",
    ".feather",
    ".ipc",
    ".arrow",
}
_IMAGE_EXTENSIONS = {".png", ".jpg", ".jpeg", ".webp"}

SPEC = ToolSpec(
    name="read_file",
    description=(
        "Read a workspace file safely, with line-range support for text files and "
        "compact summarization for tabular data such as CSV, TSV, and Excel, plus "
        "text extraction for DOCX files and text extraction and metadata for PDF "
        "files."
    ),
    parameters_schema={
        "type": "object",
        "properties": {
            "path": {
                "type": "string",
                "description": (
                    "File path relative to the workspace root "
                    "(or absolute within workspace)."
                ),
            },
            "start_line": {
                "type": "integer",
                "description": "1-based starting line number. Defaults to 1.",
            },
            "max_lines": {
                "type": "integer",
                "description": "Maximum number of lines to return. Defaults to 200.",
            },
            "encoding": {
                "type": "string",
                "description": "Text encoding to use. Defaults to 'auto'.",
                "default": "auto",
            },
        },
        "required": ["path"],
        "additionalProperties": False,
    },
)


def handle(arguments: dict[str, Any], context: ToolContext) -> dict[str, Any]:
    del context
    path_value = arguments.get("path", "")
    if not isinstance(path_value, str) or not path_value.strip():
        return {"error": "'path' must be a non-empty string."}

    root = Path.cwd().resolve()
    start_line = normalize_positive_int(arguments.get("start_line"), default=1)
    max_lines = normalize_positive_int(
        arguments.get("max_lines"), default=DEFAULT_MAX_LINES
    )
    encoding = arguments.get("encoding", "auto")

    try:
        target_path = resolve_safe_path(root, path_value)
        if not target_path.exists():
            return {"error": f"path not found: {target_path}"}
        if not target_path.is_file():
            return {"error": f"path is not a file: {target_path}"}

        suffix = target_path.suffix.lower()
        if suffix in _IMAGE_EXTENSIONS:
            return {
                "error": (
                    f"image file is not supported by read_file: {target_path.name}; "
                    "use read_image instead"
                )
            }
        if suffix in {".xlsx", ".xls"}:
            return _handle_excel_workbook(root, target_path)
        if suffix in _TABULAR_EXTENSIONS:
            return _handle_tabular_file(root, target_path, suffix)
        if suffix == ".docx":
            return _handle_docx_file(root, target_path)
        if suffix == ".pdf":
            return _handle_pdf_file(root, target_path)

        selected_lines: list[str] = []
        line_count = 0
        with open_text_file(target_path, encoding=str(encoding)) as text_handle:
            for line_count, line in enumerate(text_handle, start=1):
                if line_count < start_line:
                    continue
                if len(selected_lines) < max_lines:
                    selected_lines.append(line)

        selected = "".join(selected_lines)
        bounded_content, content_truncated = bound_output(
            selected, limit=MAX_READ_FILE_OUTPUT_CHARS
        )
        returned_start_line = start_line if selected_lines else 0
        returned_end_line = (
            returned_start_line + len(selected_lines) - 1 if selected_lines else 0
        )
        has_more_lines = returned_end_line < line_count if returned_end_line else False

        result: dict[str, Any] = {
            "path": relative_workspace_path(root, target_path),
            "start_line": returned_start_line,
            "end_line": returned_end_line,
            "total_lines": line_count,
            "content": bounded_content,
            "has_more_lines": has_more_lines,
        }
        if line_count == 0:
            result["empty"] = True
        if start_line > 1 or has_more_lines:
            result["windowed"] = True
        if content_truncated:
            result["content_truncated"] = True
        return result
    except Exception as exc:
        return {"error": bound_output(str(exc))[0]}


def _handle_tabular_file(root: Path, target_path: Path, suffix: str) -> dict[str, Any]:
    dataframe = _read_tabular_dataframe(target_path, suffix)
    result = _summarize_dataframe(dataframe)
    result["path"] = relative_workspace_path(root, target_path)
    return result


def _handle_excel_workbook(root: Path, target_path: Path) -> dict[str, Any]:
    workbook = _read_excel_workbook(target_path)
    sheets: list[dict[str, Any]] = []

    for sheet_name, dataframe in workbook.items():
        sheet_result = _summarize_dataframe(dataframe)
        sheet_result["name"] = sheet_name
        sheets.append(sheet_result)

    return {
        "path": relative_workspace_path(root, target_path),
        "sheet_count": len(sheets),
        "sheets": sheets,
    }


def _handle_docx_file(root: Path, target_path: Path) -> dict[str, Any]:
    full_text = _extract_docx_text(target_path)

    result: dict[str, Any] = {
        "path": relative_workspace_path(root, target_path),
        "content": full_text,
    }
    return result


def _extract_docx_text(path: Path) -> str:
    from docx import Document
    from docx.opc.exceptions import PackageNotFoundError

    try:
        document = Document(str(path))
    except (BadZipFile, PackageNotFoundError) as exc:
        raise ValueError(
            f"unable to read docx file: {path.name} is unreadable or corrupt"
        ) from exc

    parts: list[str] = []

    for block in document.iter_inner_content():
        from docx.table import Table
        from docx.text.paragraph import Paragraph

        if isinstance(block, Paragraph):
            parts.append(block.text)
        elif isinstance(block, Table):
            for row in block.rows:
                parts.append("\t".join(cell.text for cell in row.cells))

    for section in document.sections:
        for header_footer in (section.header, section.footer):
            if header_footer.is_linked_to_previous:
                continue
            for paragraph in header_footer.paragraphs:
                text = paragraph.text.strip()
                if text:
                    parts.append(text)

    return "\n".join(parts)


def _summarize_dataframe(dataframe: Any) -> dict[str, Any]:
    schema = dataframe.schema
    rows = dataframe.height
    columns = dataframe.width
    null_counts = {
        column_name: int(null_count)
        for column_name, null_count in zip(
            dataframe.columns, dataframe.null_count().row(0), strict=True
        )
    }
    kind_by_column = {
        column_name: _column_kind(dataframe, column_name, dtype)
        for column_name, dtype in schema.items()
    }

    schema_lines = [
        _summarize_tabular_column(
            dataframe,
            column_name,
            dtype,
            kind=kind_by_column[column_name],
            null_count=null_counts[column_name],
            row_count=rows,
        )
        for column_name, dtype in schema.items()
    ]
    schema_text, schema_truncated = bound_output(
        "\n".join(f"- {line}" for line in schema_lines),
        limit=MAX_TABULAR_SCHEMA_CHARS,
    )
    preview_text, preview_truncated = bound_output(
        _render_preview_csv(dataframe),
        limit=MAX_TABULAR_PREVIEW_CHARS,
    )

    result: dict[str, Any] = {
        "shape": {
            "rows": rows,
            "columns": columns,
        },
        "summary": _tabular_dataset_summary(
            rows=rows,
            columns=columns,
            kind_by_column=kind_by_column,
            null_counts=null_counts,
        ),
        "schema": schema_text,
        "preview": preview_text,
    }
    if schema_truncated:
        result["schema_truncated"] = True
    if preview_truncated:
        result["preview_truncated"] = True
    return result


def _read_tabular_dataframe(path: Path, suffix: str) -> Any:
    import polars as pl

    if suffix == ".csv":
        return _read_delimited_dataframe(path, separator=",")
    if suffix == ".tsv":
        return _read_delimited_dataframe(path, separator="\t")
    if suffix == ".parquet":
        return pl.read_parquet(path)
    if suffix in {".feather", ".ipc", ".arrow"}:
        return pl.read_ipc(path)
    raise ValueError(f"unsupported tabular file type: {suffix}")


def _read_excel_workbook(path: Path) -> dict[str, Any]:
    import polars as pl

    workbook = pl.read_excel(path, sheet_id=0, infer_schema_length=10_000)
    if isinstance(workbook, dict):
        return {
            sheet_name: _coerce_temporal_columns(dataframe)
            for sheet_name, dataframe in workbook.items()
        }
    return {"Sheet1": _coerce_temporal_columns(workbook)}


def _read_delimited_dataframe(path: Path, *, separator: str) -> Any:
    import polars as pl

    raw_bytes = path.read_bytes()
    if b"\r" in raw_bytes and b"\n" not in raw_bytes:
        raw_bytes = raw_bytes.replace(b"\r", b"\n")
        dataframe = pl.read_csv(
            io.BytesIO(raw_bytes),
            separator=separator,
            infer_schema_length=10_000,
        )
    else:
        dataframe = pl.read_csv(
            path,
            separator=separator,
            infer_schema_length=10_000,
        )
    return _coerce_temporal_columns(dataframe)


def _coerce_temporal_columns(dataframe: Any) -> Any:
    import polars as pl

    temporal_formats: tuple[tuple[Any, str], ...] = (
        (pl.Date, "%Y-%m-%d"),
        (pl.Date, "%m/%d/%Y"),
        (pl.Datetime, "%Y-%m-%d %H:%M:%S"),
        (pl.Datetime, "%Y-%m-%dT%H:%M:%S"),
    )
    expressions: list[Any] = []

    for column_name, dtype in dataframe.schema.items():
        if dtype != pl.String:
            continue

        series = dataframe[column_name]
        non_null = series.drop_nulls()
        if non_null.len() == 0:
            continue

        cleaned = series.str.strip_chars()
        for temporal_type, format_string in temporal_formats:
            parsed = cleaned.str.strptime(
                temporal_type,
                format=format_string,
                strict=False,
            )
            if parsed.drop_nulls().len() == non_null.len():
                expressions.append(parsed.alias(column_name))
                break

    if not expressions:
        return dataframe
    return dataframe.with_columns(expressions)


def _is_categorical_column(dataframe: Any, name: str, dtype: Any) -> bool:
    import polars as pl

    if dtype in {pl.Categorical, pl.Enum}:
        return True
    if dtype != pl.String or dataframe.height == 0:
        return False

    non_null = dataframe[name].drop_nulls()
    if non_null.len() == 0:
        return False
    unique_ratio = non_null.n_unique() / non_null.len()
    return unique_ratio <= 0.2 and non_null.n_unique() <= 100


def _column_kind(dataframe: Any, name: str, dtype: Any) -> str:
    import polars as pl

    if dtype.is_numeric():
        return "numeric"
    if dtype in {pl.Date, pl.Datetime, pl.Time}:
        return "datetime"
    if dtype == pl.Boolean:
        return "boolean"
    if _is_categorical_column(dataframe, name, dtype):
        return "categorical"
    if dtype in {pl.String, pl.Categorical, pl.Enum}:
        return "text"
    return "other"


def _tabular_dataset_summary(
    *,
    rows: int,
    columns: int,
    kind_by_column: dict[str, str],
    null_counts: dict[str, int],
) -> str:
    kind_order = ["numeric", "datetime", "categorical", "text", "boolean", "other"]
    kind_counts = {
        kind: sum(1 for value in kind_by_column.values() if value == kind)
        for kind in kind_order
    }
    mix_parts = [f"{count} {kind}" for kind, count in kind_counts.items() if count > 0]
    total_nulls = sum(null_counts.values())
    missing_summary = (
        "no missing values"
        if total_nulls == 0
        else f"{total_nulls} missing values across {sum(1 for count in null_counts.values() if count > 0)} columns"
    )

    parts = [f"{rows} rows x {columns} columns"]
    if mix_parts:
        parts.append("column mix: " + ", ".join(mix_parts))
    parts.append(f"missing values: {missing_summary}")
    return "; ".join(parts) + "."


def _summarize_tabular_column(
    dataframe: Any,
    column_name: str,
    dtype: Any,
    *,
    kind: str,
    null_count: int,
    row_count: int,
) -> str:
    series = dataframe[column_name]
    parts = [f"{column_name}: {dtype}"]
    if kind != "other":
        parts.append(f"kind={kind}")
    if null_count:
        null_ratio = (null_count / row_count) * 100 if row_count else 0.0
        parts.append(f"nulls={null_count} ({_format_number(null_ratio)}%)")

    non_null = series.drop_nulls()
    if non_null.len() == 0:
        parts.append("all values null")
        return "; ".join(parts)

    if kind == "numeric":
        parts.extend(
            [
                f"min={_format_scalar(non_null.min())}",
                f"max={_format_scalar(non_null.max())}",
                f"mean={_format_scalar(non_null.mean())}",
            ]
        )
        return "; ".join(parts)

    if kind == "datetime":
        parts.append(
            f"range={_format_scalar(non_null.min())}..{_format_scalar(non_null.max())}"
        )
        return "; ".join(parts)

    if kind == "boolean":
        parts.append(
            "values=" + ", ".join(_ordered_examples(non_null.to_list(), limit=2))
        )
        return "; ".join(parts)

    if kind == "categorical":
        distinct = non_null.n_unique()
        examples = _top_examples(non_null, limit=5)
        parts.append(f"distinct={distinct}")
        if examples:
            parts.append("examples=" + ", ".join(examples))
        return "; ".join(parts)

    if kind == "text":
        examples = _ordered_examples(non_null.to_list(), limit=3)
        if examples:
            parts.append("examples=" + ", ".join(examples))
        return "; ".join(parts)

    return "; ".join(parts)


def _render_preview_csv(dataframe: Any) -> str:
    preview = dataframe.head(DATAFRAME_PREVIEW_ROWS)
    if preview.height == 0:
        return ""

    buffer = io.StringIO()
    writer = csv.DictWriter(
        buffer,
        fieldnames=preview.columns,
        lineterminator="\n",
    )
    writer.writeheader()
    for row in preview.to_dicts():
        writer.writerow(
            {
                column_name: _format_preview_value(row.get(column_name))
                for column_name in preview.columns
            }
        )
    return buffer.getvalue()


def _top_examples(series: Any, *, limit: int) -> list[str]:
    values = series.value_counts(sort=True)[series.name].to_list()
    return _ordered_examples(values, limit=limit)


def _ordered_examples(values: list[Any], *, limit: int) -> list[str]:
    examples: list[str] = []
    seen: set[str] = set()
    for value in values:
        rendered = _format_scalar(value)
        if rendered in seen:
            continue
        seen.add(rendered)
        examples.append(rendered)
        if len(examples) >= limit:
            break
    return examples


def _format_preview_value(value: Any) -> str:
    if value is None:
        return ""
    if isinstance(value, (dict, list)):
        return str(value)
    return _format_scalar(value)


def _format_scalar(value: Any) -> str:
    if value is None:
        return "null"
    if isinstance(value, datetime):
        return value.isoformat()
    if isinstance(value, (date, time)):
        return value.isoformat()
    if isinstance(value, bool):
        return str(value).lower()
    if isinstance(value, int):
        return str(value)
    if isinstance(value, float):
        return _format_number(value)
    return str(value)


def _format_number(value: float) -> str:
    if math.isnan(value):
        return "null"
    if value.is_integer():
        return str(int(value))
    return f"{value:.4f}".rstrip("0").rstrip(".")


def _handle_pdf_file(root: Path, target_path: Path) -> dict[str, Any]:
    from pypdf import PdfReader

    reader = PdfReader(str(target_path))
    page_text = [page.extract_text() or "" for page in reader.pages]
    full_text = "\n".join(page_text)

    metadata: dict[str, Any] = {"pages": len(reader.pages)}
    if reader.metadata:
        if reader.metadata.title:
            metadata["title"] = reader.metadata.title
        if reader.metadata.author:
            metadata["author"] = reader.metadata.author

    result: dict[str, Any] = {
        "path": relative_workspace_path(root, target_path),
        "content": full_text,
        "metadata": metadata,
    }
    return result



from __future__ import annotations

import json
import urllib.request

from pbi_agent.agent.system_prompt import get_system_prompt
from pbi_agent.agent.tool_runtime import ToolExecutionBatch
from pbi_agent.config import DEFAULT_GENERIC_API_URL, DEFAULT_MAX_TOKENS, Settings
from pbi_agent.models.messages import CompletedResponse, TokenUsage, ToolCall
from pbi_agent.providers.generic_provider import GenericProvider
from pbi_agent.session_store import MessageRecord
from pbi_agent.tools.types import ToolResult


def _make_settings(**overrides: object) -> Settings:
    defaults: dict[str, object] = {
        "api_key": "test-key",
        "provider": "generic",
        "generic_api_url": DEFAULT_GENERIC_API_URL,
        "model": "",
        "max_tokens": DEFAULT_MAX_TOKENS,
        "max_retries": 1,
    }
    defaults.update(overrides)
    return Settings(**defaults)


def test_generic_parse_response_normalizes_assistant_message_and_tool_calls() -> None:
    provider = GenericProvider(_make_settings())

    result = provider._parse_response(
        {
            "id": "chatcmpl_123",
            "model": "openrouter/auto",
            "usage": {
                "prompt_tokens": 21,
                "completion_tokens": 8,
                "completion_tokens_details": {"reasoning_tokens": 3},
            },
            "choices": [
                {
                    "message": {
                        "role": "assistant",
                        "content": [
                            {"type": "text", "text": "First line. "},
                            {"type": "output_text", "text": "Second line."},
                            {"type": "refusal", "refusal": "policy block"},
                        ],
                        "tool_calls": [
                            {
                                "id": "call_1",
                                "type": "function",
                                "function": {
                                    "name": "shell",
                                    "arguments": '{"command":"pwd"}',
                                },
                            },
                            {
                                "id": "call_2",
                                "type": "function",
                                "function": {
                                    "name": "init_report",
                                    "arguments": {"dest": "."},
                                },
                            },
                        ],
                    }
                }
            ],
        }
    )

    assert result.response_id == "chatcmpl_123"
    assert result.text == "First line. Second line."
    assert result.function_calls == [
        ToolCall(call_id="call_1", name="shell", arguments={"command": "pwd"}),
        ToolCall(call_id="call_2", name="init_report", arguments={"dest": "."}),
    ]
    assert result.usage.input_tokens == 21
    assert result.usage.output_tokens == 8
    assert result.usage.reasoning_tokens == 3
    assert result.usage.model == "openrouter/auto"

    assistant_message = result.provider_data["assistant_message"]
    assert assistant_message["role"] == "assistant"
    assert assistant_message["content"] == [
        {"type": "text", "text": "First line. "},
        {"type": "text", "text": "Second line."},
        {"type": "refusal", "refusal": "policy block"},
    ]
    assert [item["id"] for item in assistant_message["tool_calls"]] == [
        "call_1",
        "call_2",
    ]
    assert json.loads(assistant_message["tool_calls"][0]["function"]["arguments"]) == {
        "command": "pwd"
    }
    assert json.loads(assistant_message["tool_calls"][1]["function"]["arguments"]) == {
        "dest": "."
    }


def test_generic_request_turn_preserves_history_and_tool_results(
    monkeypatch,
    display_spy,
    make_http_response,
) -> None:
    requests: list[dict[str, object]] = []
    responses = iter(
        [
            {
                "id": "chatcmpl_1",
                "model": "openrouter/auto",
                "usage": {
                    "prompt_tokens": 10,
                    "completion_tokens": 2,
                    "completion_tokens_details": {"reasoning_tokens": 1},
                },
                "choices": [
                    {
                        "message": {
                            "role": "assistant",
                            "content": "Let me check.",
                            "tool_calls": [
                                {
                                    "id": "call_1",
                                    "type": "function",
                                    "function": {
                                        "name": "shell",
                                        "arguments": '{"command":"pwd"}',
                                    },
                                }
                            ],
                        }
                    }
                ],
            },
            {
                "id": "chatcmpl_2",
                "model": "openrouter/auto",
                "usage": {
                    "prompt_tokens": 12,
                    "completion_tokens": 5,
                    "completion_tokens_details": {"reasoning_tokens": 0},
                },
                "choices": [
                    {
                        "message": {
                            "role": "assistant",
                            "content": "You are in /workspace.",
                        }
                    }
                ],
            },
        ]
    )

    def fake_urlopen(
        request: urllib.request.Request,
        timeout: float,
    ):
        del timeout
        payload = request.data.decode("utf-8") if request.data else "{}"
        requests.append(json.loads(payload))
        return make_http_response(next(responses))

    monkeypatch.setattr(urllib.request, "urlopen", fake_urlopen)

    provider = GenericProvider(_make_settings())
    session_usage = TokenUsage()

    first = provider.request_turn(
        user_message="Where am I?",
        display=display_spy,
        session_usage=session_usage,
        turn_usage=TokenUsage(),
    )
    second = provider.request_turn(
        tool_result_items=[
            {
                "role": "tool",
                "tool_call_id": "call_1",
                "content": '{"ok": true, "result": "/workspace"}',
            }
        ],
        display=display_spy,
        session_usage=session_usage,
        turn_usage=TokenUsage(),
    )

    assert first.response_id == "chatcmpl_1"
    assert second.response_id == "chatcmpl_2"
    assert requests[0] == {
        "messages": [
            {"role": "system", "content": get_system_prompt()},
            {"role": "user", "content": "Where am I?"},
        ],
        "max_tokens": DEFAULT_MAX_TOKENS,
        "tools": provider._tools,
        "tool_choice": "auto",
    }
    assert requests[1] == {
        "messages": [
            {"role": "system", "content": get_system_prompt()},
            {"role": "user", "content": "Where am I?"},
            {
                "role": "assistant",
                "content": "Let me check.",
                "tool_calls": [
                    {
                        "id": "call_1",
                        "type": "function",
                        "function": {
                            "name": "shell",
                            "arguments": '{"command":"pwd"}',
                        },
                    }
                ],
            },
            {
                "role": "tool",
                "tool_call_id": "call_1",
                "content": '{"ok": true, "result": "/workspace"}',
            },
        ],
        "max_tokens": DEFAULT_MAX_TOKENS,
        "tools": provider._tools,
        "tool_choice": "auto",
    }
    assert display_spy.markdown_calls == ["Let me check.", "You are in /workspace."]
    assert display_spy.session_usage_snapshots[-1].input_tokens == 22
    assert display_spy.session_usage_snapshots[-1].output_tokens == 7
    assert display_spy.session_usage_snapshots[-1].reasoning_tokens == 1


def test_generic_restore_messages_reuses_persisted_history(
    monkeypatch,
    display_spy,
    make_http_response,
) -> None:
    requests: list[dict[str, object]] = []

    def fake_urlopen(
        request: urllib.request.Request,
        timeout: float,
    ):
        del timeout
        payload = request.data.decode("utf-8") if request.data else "{}"
        requests.append(json.loads(payload))
        return make_http_response(
            {
                "id": "chatcmpl_3",
                "model": "openrouter/auto",
                "usage": {
                    "prompt_tokens": 11,
                    "completion_tokens": 4,
                    "completion_tokens_details": {"reasoning_tokens": 0},
                },
                "choices": [
                    {
                        "message": {
                            "role": "assistant",
                            "content": "Follow-up answer.",
                        }
                    }
                ],
            }
        )

    monkeypatch.setattr(urllib.request, "urlopen", fake_urlopen)

    provider = GenericProvider(_make_settings())
    provider.restore_messages(
        [
            MessageRecord(
                id=1,
                session_id="session-1",
                role="user",
                content="Original question",
                created_at="2026-03-19T10:00:00+00:00",
            ),
            MessageRecord(
                id=2,
                session_id="session-1",
                role="assistant",
                content="Original answer",
                created_at="2026-03-19T10:00:01+00:00",
            ),
        ]
    )

    provider.request_turn(
        user_message="Follow-up question",
        display=display_spy,
        session_usage=TokenUsage(),
        turn_usage=TokenUsage(),
    )

    assert requests[0] == {
        "messages": [
            {"role": "system", "content": get_system_prompt()},
            {"role": "user", "content": "Original question"},
            {"role": "assistant", "content": "Original answer"},
            {"role": "user", "content": "Follow-up question"},
        ],
        "max_tokens": DEFAULT_MAX_TOKENS,
        "tools": provider._tools,
        "tool_choice": "auto",
    }
    assert display_spy.markdown_calls == ["Follow-up answer."]


def test_generic_execute_tool_calls_returns_chat_completion_tool_messages(
    monkeypatch,
    display_spy,
) -> None:
    provider = GenericProvider(_make_settings())
    response = CompletedResponse(
        response_id="chatcmpl_1",
        text="",
        function_calls=[
            ToolCall(call_id="call_1", name="shell", arguments={"command": "pwd"}),
            ToolCall(call_id="call_2", name="init_report", arguments={"dest": "."}),
        ],
    )
    batch = ToolExecutionBatch(
        results=[
            ToolResult(
                call_id="call_1",
                output_json='{"ok": true, "result": "/workspace"}',
            ),
            ToolResult(
                call_id="call_2",
                output_json=(
                    '{"ok": false, "error": {"type": "tool_execution_failed", '
                    '"message": "boom"}}'
                ),
                is_error=True,
            ),
        ],
        had_errors=True,
    )

    monkeypatch.setattr(
        "pbi_agent.providers.generic_provider._execute_tool_calls",
        lambda calls, max_workers, context=None: batch,
    )

    tool_result_items, had_errors = provider.execute_tool_calls(
        response,
        max_workers=2,
        display=display_spy,
        session_usage=TokenUsage(),
        turn_usage=TokenUsage(),
    )

    assert had_errors is True
    assert tool_result_items == [
        {
            "role": "tool",
            "tool_call_id": "call_1",
            "content": '{"ok": true, "result": "/workspace"}',
        },
        {
            "role": "tool",
            "tool_call_id": "call_2",
            "content": (
                '{"ok": false, "error": {"type": "tool_execution_failed", '
                '"message": "boom"}}'
            ),
        },
    ]
    assert display_spy.function_counts == [2]
    assert display_spy.function_results == [
        {
            "name": "shell",
            "success": True,
            "call_id": "call_1",
            "arguments": {"command": "pwd"},
        },
        {
            "name": "init_report",
            "success": False,
            "call_id": "call_2",
            "arguments": {"dest": "."},
        },
    ]
    assert display_spy.tool_group_end_count == 1


def test_generic_request_turn_retries_after_rate_limit(
    monkeypatch,
    display_spy,
    make_http_error,
    make_http_response,
) -> None:
    requests: list[dict[str, object]] = []
    waits: list[float] = []
    response_payload = {
        "id": "chatcmpl_retry",
        "model": "openrouter/my-model",
        "usage": {
            "prompt_tokens": 6,
            "completion_tokens": 4,
            "completion_tokens_details": {"reasoning_tokens": 0},
        },
        "choices": [
            {
                "message": {
                    "role": "assistant",
                    "content": "Recovered.",
                }
            }
        ],
    }

    def fake_urlopen(
        request: urllib.request.Request,
        timeout: float,
    ):
        del timeout
        payload = request.data.decode("utf-8") if request.data else "{}"
        requests.append(json.loads(payload))
        if len(requests) == 1:
            raise make_http_error(
                url=DEFAULT_GENERIC_API_URL,
                code=429,
                body='{"error":"slow down"}',
                headers={"Retry-After": "0.25"},
            )
        return make_http_response(response_payload)

    monkeypatch.setattr(urllib.request, "urlopen", fake_urlopen)
    monkeypatch.setattr(
        "pbi_agent.providers.generic_provider.time.sleep",
        lambda seconds: waits.append(seconds),
    )

    provider = GenericProvider(
        _make_settings(model="openrouter/my-model", max_retries=1)
    )
    response = provider.request_turn(
        user_message="hello",
        display=display_spy,
        session_usage=TokenUsage(),
        turn_usage=TokenUsage(),
    )

    assert response.text == "Recovered."
    assert len(requests) == 2
    assert requests[0]["model"] == "openrouter/my-model"
    assert requests[1]["model"] == "openrouter/my-model"
    assert waits == [1.25]
    assert display_spy.rate_limit_notices == [(1.25, 1, 1)]
    assert display_spy.retry_notices == [(1, 1)]



from __future__ import annotations

import json
import subprocess
from pathlib import Path

from pbi_agent.agent import tool_runtime
from pbi_agent.models.messages import ToolCall
from pbi_agent.tools import apply_patch as apply_patch_tool
from pbi_agent.tools import read_file as read_file_tool
from pbi_agent.tools import shell as shell_tool
from pbi_agent.tools.output import MAX_OUTPUT_CHARS
from pbi_agent.tools.types import ToolContext


def test_execute_tool_calls_reports_unknown_tool_error(monkeypatch) -> None:
    monkeypatch.setattr(tool_runtime, "get_tool_handler", lambda name: None)

    batch = tool_runtime.execute_tool_calls(
        [ToolCall(call_id="call_1", name="missing_tool", arguments={})],
        max_workers=1,
    )

    assert batch.had_errors is True
    assert len(batch.results) == 1
    assert json.loads(batch.results[0].output_json) == {
        "ok": False,
        "error": {
            "type": "unknown_tool",
            "message": "Tool 'missing_tool' is not registered.",
        },
        "tool": "missing_tool",
    }


def test_execute_tool_calls_rejects_invalid_json_arguments(monkeypatch) -> None:
    invoked = False

    def fake_handler(
        arguments: dict[str, object], context: ToolContext
    ) -> dict[str, object]:
        del arguments, context
        nonlocal invoked
        invoked = True
        return {}

    monkeypatch.setattr(tool_runtime, "get_tool_handler", lambda name: fake_handler)

    batch = tool_runtime.execute_tool_calls(
        [ToolCall(call_id="call_1", name="shell", arguments="[]")],
        max_workers=1,
    )

    assert invoked is False
    assert batch.had_errors is True
    assert json.loads(batch.results[0].output_json) == {
        "ok": False,
        "error": {
            "type": "invalid_arguments",
            "message": "tool arguments must decode to a JSON object",
        },
        "tool": "shell",
    }


def test_execute_tool_calls_wraps_success_payload_and_output_items(monkeypatch) -> None:
    seen: list[tuple[dict[str, object], bool]] = []

    def fake_handler(
        arguments: dict[str, object],
        context: ToolContext,
    ) -> dict[str, object]:
        seen.append((arguments, isinstance(context, ToolContext)))
        return {"echo": arguments["value"]}

    monkeypatch.setattr(tool_runtime, "get_tool_handler", lambda name: fake_handler)

    batch = tool_runtime.execute_tool_calls(
        [ToolCall(call_id="call_1", name="shell", arguments='{"value": 7}')],
        max_workers=1,
    )

    assert batch.had_errors is False
    assert seen == [({"value": 7}, True)]
    assert json.loads(batch.results[0].output_json) == {
        "ok": True,
        "result": {"echo": 7},
    }
    assert tool_runtime.to_function_call_output_items(batch.results) == [
        {
            "type": "function_call_output",
            "call_id": "call_1",
            "output": batch.results[0].output_json,
        }
    ]


def test_execute_tool_calls_wraps_handler_exceptions(monkeypatch) -> None:
    def fake_handler(
        arguments: dict[str, object],
        context: ToolContext,
    ) -> dict[str, object]:
        del arguments, context
        raise RuntimeError("boom")

    monkeypatch.setattr(tool_runtime, "get_tool_handler", lambda name: fake_handler)

    batch = tool_runtime.execute_tool_calls(
        [ToolCall(call_id="call_1", name="shell", arguments={})],
        max_workers=1,
    )

    assert batch.had_errors is True
    assert json.loads(batch.results[0].output_json) == {
        "ok": False,
        "error": {
            "type": "tool_execution_failed",
            "message": "boom",
        },
        "tool": "shell",
    }


def test_execute_tool_calls_serializes_truncated_shell_output(
    tmp_path: Path,
    monkeypatch,
) -> None:
    monkeypatch.chdir(tmp_path)
    stdout = f"stdout-start-{'a' * (MAX_OUTPUT_CHARS + 50)}-stdout-end"
    stderr = f"stderr-start-{'b' * (MAX_OUTPUT_CHARS + 50)}-stderr-end"

    def fake_run(*args: object, **kwargs: object) -> subprocess.CompletedProcess[bytes]:
        del args, kwargs
        return subprocess.CompletedProcess(
            args="python",
            returncode=0,
            stdout=stdout.encode("utf-8"),
            stderr=stderr.encode("utf-8"),
        )

    monkeypatch.setattr(shell_tool.subprocess, "run", fake_run)
    monkeypatch.setattr(
        tool_runtime,
        "get_tool_handler",
        lambda name: shell_tool.handle if name == "shell" else None,
    )

    batch = tool_runtime.execute_tool_calls(
        [ToolCall(call_id="call_1", name="shell", arguments={"command": "python"})],
        max_workers=1,
    )

    payload = json.loads(batch.results[0].output_json)
    result = payload["result"]

    assert payload["ok"] is True
    assert result["exit_code"] == 0
    assert result["stdout_truncated"] is True
    assert result["stderr_truncated"] is True
    assert len(result["stdout"]) <= MAX_OUTPUT_CHARS
    assert len(result["stderr"]) <= MAX_OUTPUT_CHARS
    assert result["stdout"].startswith("stdout-start-")
    assert result["stdout"].endswith("-stdout-end")
    assert result["stderr"].startswith("stderr-start-")
    assert result["stderr"].endswith("-stderr-end")
    assert "chars omitted" in result["stdout"]
    assert "chars omitted" in result["stderr"]


def test_execute_tool_calls_serializes_truncated_apply_patch_error(
    monkeypatch,
) -> None:
    def raise_long_error(path: Path, diff: str | None) -> None:
        del path, diff
        raise ValueError(f"start-{'x' * (MAX_OUTPUT_CHARS + 200)}-end")

    monkeypatch.setattr(apply_patch_tool, "_create_file", raise_long_error)
    monkeypatch.setattr(
        tool_runtime,
        "get_tool_handler",
        lambda name: apply_patch_tool.handle if name == "apply_patch" else None,
    )

    batch = tool_runtime.execute_tool_calls(
        [
            ToolCall(
                call_id="call_1",
                name="apply_patch",
                arguments={
                    "operation_type": "create_file",
                    "path": "notes/example.txt",
                    "diff": "+hello",
                },
            )
        ],
        max_workers=1,
    )

    payload = json.loads(batch.results[0].output_json)
    result = payload["result"]

    assert payload["ok"] is True
    assert result["status"] == "failed"
    assert len(result["error"]) <= MAX_OUTPUT_CHARS
    assert result["error"].startswith("start-")
    assert result["error"].endswith("-end")
    assert "chars omitted" in result["error"]


def test_execute_tool_calls_serializes_tabular_read_file_output(
    tmp_path: Path,
    monkeypatch,
) -> None:
    monkeypatch.chdir(tmp_path)
    (tmp_path / "dataset.csv").write_text(
        "ordered_at,value\n2025-01-01,1\n2025-01-02,2\n",
        encoding="utf-8",
    )
    monkeypatch.setattr(
        tool_runtime,
        "get_tool_handler",
        lambda name: read_file_tool.handle if name == "read_file" else None,
    )

    batch = tool_runtime.execute_tool_calls(
        [
            ToolCall(
                call_id="call_1",
                name="read_file",
                arguments={"path": "dataset.csv"},
            )
        ],
        max_workers=1,
    )

    payload = json.loads(batch.results[0].output_json)

    assert payload["ok"] is True
    assert payload["result"]["shape"] == {"rows": 2, "columns": 2}
    assert (
        payload["result"]["preview"] == "ordered_at,value\n2025-01-01,1\n2025-01-02,2\n"
    )


def test_execute_tool_calls_passes_runtime_context_to_handler(monkeypatch) -> None:
    captured_context: list[ToolContext] = []

    def fake_handler(
        arguments: dict[str, object],
        context: ToolContext,
    ) -> dict[str, object]:
        del arguments
        captured_context.append(context)
        return {"ok": True}

    monkeypatch.setattr(tool_runtime, "get_tool_handler", lambda name: fake_handler)

    batch = tool_runtime.execute_tool_calls(
        [ToolCall(call_id="call_1", name="shell", arguments={})],
        max_workers=1,
        context=ToolContext(sub_agent_depth=1),
    )

    assert batch.had_errors is False
    assert len(captured_context) == 1
    assert captured_context[0].sub_agent_depth == 1



from __future__ import annotations

import json
import urllib.request

import pytest

from pbi_agent.agent.system_prompt import get_system_prompt
from pbi_agent.agent.tool_runtime import ToolExecutionBatch
from pbi_agent.config import (
    DEFAULT_ANTHROPIC_MODEL,
    DEFAULT_MAX_TOKENS,
    Settings,
)
from pbi_agent.models.messages import (
    CompletedResponse,
    ImageAttachment,
    TokenUsage,
    UserTurnInput,
)
from pbi_agent.providers.anthropic_provider import ANTHROPIC_API_URL, AnthropicProvider
from pbi_agent.session_store import MessageRecord
from pbi_agent.tools.types import ToolResult


def _make_settings(**overrides: object) -> Settings:
    defaults: dict[str, object] = {
        "api_key": "test-key",
        "provider": "anthropic",
        "model": DEFAULT_ANTHROPIC_MODEL,
        "max_tokens": DEFAULT_MAX_TOKENS,
        "reasoning_effort": "xhigh",
        "max_retries": 1,
    }
    defaults.update(overrides)
    return Settings(**defaults)


def test_anthropic_parse_response_extracts_cache_usage_and_tool_calls() -> None:
    provider = AnthropicProvider(_make_settings())

    result = provider._parse_response(
        {
            "id": "msg_123",
            "content": [
                {"type": "thinking", "thinking": "Inspecting the report."},
                {"type": "redacted_thinking"},
                {"type": "text", "text": "I will inspect the model."},
                {
                    "type": "tool_use",
                    "id": "toolu_1",
                    "name": "shell",
                    "input": {"command": "pwd"},
                },
                {"type": "text", "text": "One command should be enough."},
            ],
            "usage": {
                "input_tokens": 100,
                "output_tokens": 30,
                "cache_read_input_tokens": 20,
                "cache_creation_input_tokens": 15,
                "cache_creation": {"ephemeral_1h_input_tokens": 5},
            },
        }
    )

    assert result.response_id == "msg_123"
    assert result.text == "I will inspect the model.\n\nOne command should be enough."
    assert len(result.function_calls) == 1
    assert result.function_calls[0].call_id == "toolu_1"
    assert result.function_calls[0].name == "shell"
    assert result.function_calls[0].arguments == {"command": "pwd"}
    assert result.usage.input_tokens == 135
    assert result.usage.cached_input_tokens == 20
    assert result.usage.cache_write_tokens == 10
    assert result.usage.cache_write_1h_tokens == 5
    assert result.usage.output_tokens == 30
    assert result.provider_data["thinking_parts"] == ["Inspecting the report."]
    assert result.provider_data["has_redacted_thinking"] is True
    assert len(result.provider_data["content_blocks"]) == 5


def test_anthropic_request_turn_preserves_history_and_wraps_tool_results(
    monkeypatch,
    display_spy,
    make_http_response,
) -> None:
    requests: list[dict[str, object]] = []
    first_content = [
        {"type": "thinking", "thinking": "Checking the workspace."},
        {"type": "redacted_thinking"},
        {
            "type": "tool_use",
            "id": "toolu_1",
            "name": "shell",
            "input": {"command": "pwd"},
        },
    ]
    responses = iter(
        [
            {
                "id": "msg_1",
                "content": first_content,
                "usage": {
                    "input_tokens": 10,
                    "output_tokens": 2,
                },
            },
            {
                "id": "msg_2",
                "content": [{"type": "text", "text": "You are in /workspace."}],
                "usage": {
                    "input_tokens": 12,
                    "output_tokens": 4,
                    "cache_read_input_tokens": 3,
                    "cache_creation_input_tokens": 5,
                    "cache_creation": {"ephemeral_1h_input_tokens": 2},
                },
            },
        ]
    )

    def fake_urlopen(
        request: urllib.request.Request,
        timeout: float,
    ):
        del timeout
        payload = request.data.decode("utf-8") if request.data else "{}"
        requests.append(json.loads(payload))
        return make_http_response(next(responses))

    monkeypatch.setattr(urllib.request, "urlopen", fake_urlopen)

    provider = AnthropicProvider(_make_settings())
    session_usage = TokenUsage()

    first = provider.request_turn(
        user_message="Where am I?",
        display=display_spy,
        session_usage=session_usage,
        turn_usage=TokenUsage(),
    )
    second = provider.request_turn(
        tool_result_items=[
            {
                "type": "tool_result",
                "tool_use_id": "toolu_1",
                "content": '{"ok": true, "result": "/workspace"}',
            }
        ],
        display=display_spy,
        session_usage=session_usage,
        turn_usage=TokenUsage(),
    )

    assert first.response_id == "msg_1"
    assert second.response_id == "msg_2"
    assert requests[0]["model"] == DEFAULT_ANTHROPIC_MODEL
    assert requests[0]["max_tokens"] == DEFAULT_MAX_TOKENS
    assert requests[0]["thinking"] == {"type": "adaptive"}
    assert requests[0]["output_config"] == {"effort": "max"}
    assert requests[0]["system"] == get_system_prompt()
    assert requests[0]["messages"] == [
        {
            "role": "user",
            "content": [{"type": "text", "text": "Where am I?"}],
        }
    ]
    assert requests[1]["messages"] == [
        {
            "role": "user",
            "content": [{"type": "text", "text": "Where am I?"}],
        },
        {"role": "assistant", "content": first_content},
        {
            "role": "user",
            "content": [
                {
                    "type": "tool_result",
                    "tool_use_id": "toolu_1",
                    "content": '{"ok": true, "result": "/workspace"}',
                }
            ],
        },
    ]
    assert display_spy.thinking_calls[0]["text"] == "Checking the workspace."
    assert display_spy.redacted_thinking_calls == 1
    assert display_spy.markdown_calls == ["You are in /workspace."]
    assert display_spy.session_usage_snapshots[-1].input_tokens == 30
    assert display_spy.session_usage_snapshots[-1].output_tokens == 6
    assert display_spy.session_usage_snapshots[-1].cached_input_tokens == 3
    assert display_spy.session_usage_snapshots[-1].cache_write_tokens == 3
    assert display_spy.session_usage_snapshots[-1].cache_write_1h_tokens == 2


def test_anthropic_request_turn_omits_adaptive_thinking_for_haiku_models(
    monkeypatch,
    display_spy,
    make_http_response,
) -> None:
    requests: list[dict[str, object]] = []

    def fake_urlopen(
        request: urllib.request.Request,
        timeout: float,
    ):
        del timeout
        payload = request.data.decode("utf-8") if request.data else "{}"
        requests.append(json.loads(payload))
        return make_http_response(
            {
                "id": "msg_haiku",
                "content": [{"type": "text", "text": "Bitcoin is $1."}],
                "usage": {
                    "input_tokens": 8,
                    "output_tokens": 3,
                },
            }
        )

    monkeypatch.setattr(urllib.request, "urlopen", fake_urlopen)

    provider = AnthropicProvider(
        _make_settings(model="claude-haiku-4-5", reasoning_effort="low")
    )
    provider.request_turn(
        user_message="check bitcoin price",
        display=display_spy,
        session_usage=TokenUsage(),
        turn_usage=TokenUsage(),
    )

    assert "thinking" not in requests[0]
    assert "output_config" not in requests[0]


def test_anthropic_restore_messages_reuses_persisted_history(
    monkeypatch,
    display_spy,
    make_http_response,
) -> None:
    requests: list[dict[str, object]] = []

    def fake_urlopen(
        request: urllib.request.Request,
        timeout: float,
    ):
        del timeout
        payload = request.data.decode("utf-8") if request.data else "{}"
        requests.append(json.loads(payload))
        return make_http_response(
            {
                "id": "msg_3",
                "content": [{"type": "text", "text": "Follow-up answer."}],
                "usage": {
                    "input_tokens": 8,
                    "output_tokens": 3,
                },
            }
        )

    monkeypatch.setattr(urllib.request, "urlopen", fake_urlopen)

    provider = AnthropicProvider(_make_settings())
    provider.restore_messages(
        [
            MessageRecord(
                id=1,
                session_id="session-1",
                role="user",
                content="Original question",
                created_at="2026-03-19T10:00:00+00:00",
            ),
            MessageRecord(
                id=2,
                session_id="session-1",
                role="assistant",
                content="Original answer",
                created_at="2026-03-19T10:00:01+00:00",
            ),
        ]
    )

    provider.request_turn(
        user_message="Follow-up question",
        display=display_spy,
        session_usage=TokenUsage(),
        turn_usage=TokenUsage(),
    )

    assert requests[0]["messages"] == [
        {
            "role": "user",
            "content": [{"type": "text", "text": "Original question"}],
        },
        {
            "role": "assistant",
            "content": [{"type": "text", "text": "Original answer"}],
        },
        {
            "role": "user",
            "content": [{"type": "text", "text": "Follow-up question"}],
        },
    ]


def test_anthropic_execute_tool_calls_returns_tool_result_blocks(
    monkeypatch,
    display_spy,
) -> None:
    provider = AnthropicProvider(_make_settings())
    response = CompletedResponse(
        response_id="msg_1",
        text="",
        provider_data={
            "content_blocks": [
                {
                    "type": "tool_use",
                    "id": "toolu_1",
                    "name": "shell",
                    "input": {"command": "pwd"},
                },
                {
                    "type": "tool_use",
                    "id": "toolu_2",
                    "name": "init_report",
                    "input": {"dest": "."},
                },
            ]
        },
    )
    batch = ToolExecutionBatch(
        results=[
            ToolResult(
                call_id="toolu_1",
                output_json='{"ok": true, "result": "/workspace"}',
            ),
            ToolResult(
                call_id="toolu_2",
                output_json=(
                    '{"ok": false, "error": {"type": "tool_execution_failed", '
                    '"message": "boom"}}'
                ),
                is_error=True,
            ),
        ],
        had_errors=True,
    )

    monkeypatch.setattr(
        "pbi_agent.providers.anthropic_provider._execute_tool_calls",
        lambda calls, max_workers, context=None: batch,
    )

    tool_result_items, had_errors = provider.execute_tool_calls(
        response,
        max_workers=2,
        display=display_spy,
        session_usage=TokenUsage(),
        turn_usage=TokenUsage(),
    )

    assert had_errors is True
    assert tool_result_items == [
        {
            "type": "tool_result",
            "tool_use_id": "toolu_1",
            "content": '{"ok": true, "result": "/workspace"}',
        },
        {
            "type": "tool_result",
            "tool_use_id": "toolu_2",
            "content": (
                '{"ok": false, "error": {"type": "tool_execution_failed", '
                '"message": "boom"}}'
            ),
            "is_error": True,
        },
    ]
    assert display_spy.function_counts == [2]
    assert display_spy.function_results == [
        {
            "name": "shell",
            "success": True,
            "call_id": "toolu_1",
            "arguments": {"command": "pwd"},
        },
        {
            "name": "init_report",
            "success": False,
            "call_id": "toolu_2",
            "arguments": {"dest": "."},
        },
    ]
    assert display_spy.tool_group_end_count == 1


def test_anthropic_request_turn_retries_when_api_is_overloaded(
    monkeypatch,
    display_spy,
    make_http_error,
    make_http_response,
) -> None:
    waits: list[float] = []
    requests: list[dict[str, object]] = []
    response_payload = {
        "id": "msg_retry",
        "content": [{"type": "text", "text": "Recovered."}],
        "usage": {"input_tokens": 5, "output_tokens": 3},
    }

    def fake_urlopen(
        request: urllib.request.Request,
        timeout: float,
    ):
        del timeout
        payload = request.data.decode("utf-8") if request.data else "{}"
        requests.append(json.loads(payload))
        if len(requests) == 1:
            raise make_http_error(
                url=ANTHROPIC_API_URL,
                code=529,
                body='{"error":"overloaded"}',
            )
        return make_http_response(response_payload)

    monkeypatch.setattr(urllib.request, "urlopen", fake_urlopen)
    monkeypatch.setattr(
        "pbi_agent.providers.anthropic_provider.time.sleep",
        lambda seconds: waits.append(seconds),
    )

    provider = AnthropicProvider(_make_settings(max_retries=1))
    response = provider.request_turn(
        user_message="hello",
        display=display_spy,
        session_usage=TokenUsage(),
        turn_usage=TokenUsage(),
    )

    assert response.text == "Recovered."
    assert len(requests) == 2
    assert waits == [3.0]
    assert display_spy.overload_notices == [(3.0, 1, 1)]
    assert display_spy.rate_limit_notices == []
    assert display_spy.retry_notices == [(1, 1)]


def test_anthropic_request_turn_preserves_error_type_and_request_id(
    monkeypatch,
    display_spy,
    make_http_error,
) -> None:
    def fake_urlopen(
        request: urllib.request.Request,
        timeout: float,
    ):
        del request, timeout
        raise make_http_error(
            url=ANTHROPIC_API_URL,
            code=404,
            body=(
                '{"type":"error","error":{"type":"not_found_error",'
                '"message":"The requested resource could not be found."},'
                '"request_id":"req_404"}'
            ),
        )

    monkeypatch.setattr(urllib.request, "urlopen", fake_urlopen)

    provider = AnthropicProvider(_make_settings(max_retries=0))

    with pytest.raises(RuntimeError) as exc_info:
        provider.request_turn(
            user_message="hello",
            display=display_spy,
            session_usage=TokenUsage(),
            turn_usage=TokenUsage(),
        )

    assert str(exc_info.value) == (
        'Anthropic API error 404: {"error": {"message": "The requested resource '
        'could not be found.", "type": "not_found_error"}, "request_id": '
        '"req_404", "status": 404, "type": "error"}'
    )


def test_anthropic_request_turn_uses_request_id_header_for_413_errors(
    monkeypatch,
    display_spy,
    make_http_error,
) -> None:
    def fake_urlopen(
        request: urllib.request.Request,
        timeout: float,
    ):
        del request, timeout
        raise make_http_error(
            url=ANTHROPIC_API_URL,
            code=413,
            body="<html>Request Entity Too Large</html>",
            headers={"request-id": "req_413"},
        )

    monkeypatch.setattr(urllib.request, "urlopen", fake_urlopen)

    provider = AnthropicProvider(_make_settings(max_retries=0))

    with pytest.raises(RuntimeError) as exc_info:
        provider.request_turn(
            user_message="hello",
            display=display_spy,
            session_usage=TokenUsage(),
            turn_usage=TokenUsage(),
        )

    assert str(exc_info.value) == (
        'Anthropic API error 413: {"error": {"message": "Request exceeds the '
        'maximum allowed number of bytes.", "type": "request_too_large"}, '
        '"request_id": "req_413", "status": 413, "type": "error"}'
    )


def test_anthropic_request_turn_serializes_user_input_images(
    monkeypatch,
    display_spy,
    make_http_response,
) -> None:
    requests: list[dict[str, object]] = []

    def fake_urlopen(
        request: urllib.request.Request,
        timeout: float,
    ):
        del timeout
        payload = request.data.decode("utf-8") if request.data else "{}"
        requests.append(json.loads(payload))
        return make_http_response(
            {
                "id": "msg_1",
                "content": [{"type": "text", "text": "done"}],
                "usage": {"input_tokens": 1, "output_tokens": 1},
            }
        )

    monkeypatch.setattr(urllib.request, "urlopen", fake_urlopen)

    provider = AnthropicProvider(_make_settings())
    provider.request_turn(
        user_input=UserTurnInput(
            text="Describe this image.",
            images=[
                ImageAttachment(
                    path="chart.png",
                    mime_type="image/png",
                    data_base64="QUJDRA==",
                )
            ],
        ),
        display=display_spy,
        session_usage=TokenUsage(),
        turn_usage=TokenUsage(),
    )

    assert requests[0]["messages"][0]["content"] == [
        {
            "type": "image",
            "source": {
                "type": "base64",
                "media_type": "image/png",
                "data": "QUJDRA==",
            },
        },
        {"type": "text", "text": "Describe this image."},
    ]


def test_anthropic_execute_tool_calls_serializes_image_attachments(
    monkeypatch,
    display_spy,
) -> None:
    provider = AnthropicProvider(_make_settings())
    response = CompletedResponse(
        response_id="msg_1",
        text="",
        provider_data={
            "content_blocks": [
                {
                    "type": "tool_use",
                    "id": "toolu_1",
                    "name": "read_image",
                    "input": {"path": "chart.png"},
                }
            ]
        },
    )
    batch = ToolExecutionBatch(
        results=[
            ToolResult(
                call_id="toolu_1",
                output_json='{"ok": true, "result": {"path": "chart.png"}}',
                attachments=[
                    ImageAttachment(
                        path="chart.png",
                        mime_type="image/png",
                        data_base64="QUJDRA==",
                    )
                ],
            )
        ],
        had_errors=False,
    )

    monkeypatch.setattr(
        "pbi_agent.providers.anthropic_provider._execute_tool_calls",
        lambda calls, max_workers, context=None: batch,
    )

    tool_result_items, had_errors = provider.execute_tool_calls(
        response,
        max_workers=1,
        display=display_spy,
        session_usage=TokenUsage(),
        turn_usage=TokenUsage(),
    )

    assert had_errors is False
    assert tool_result_items == [
        {
            "type": "tool_result",
            "tool_use_id": "toolu_1",
            "content": [
                {
                    "type": "image",
                    "source": {
                        "type": "base64",
                        "media_type": "image/png",
                        "data": "QUJDRA==",
                    },
                },
                {
                    "type": "text",
                    "text": '{"ok": true, "result": {"path": "chart.png"}}',
                },
            ],
        }
    ]


def test_anthropic_web_search_tool_included_when_enabled() -> None:
    provider = AnthropicProvider(_make_settings(web_search=True))
    web_tool = {"type": "web_search_20260209", "name": "web_search"}
    assert web_tool in provider._tools


def test_anthropic_web_search_tool_uses_direct_callers_for_haiku_models() -> None:
    provider = AnthropicProvider(
        _make_settings(web_search=True, model="claude-haiku-4-5")
    )
    assert {
        "type": "web_search_20260209",
        "name": "web_search",
        "allowed_callers": ["direct"],
    } in provider._tools


def test_anthropic_web_search_tool_excluded_when_disabled() -> None:
    provider = AnthropicProvider(_make_settings(web_search=False))
    tool_types = [t.get("type") for t in provider._tools]
    assert "web_search_20260209" not in tool_types


def test_anthropic_parse_response_extracts_web_search_sources() -> None:
    provider = AnthropicProvider(_make_settings())

    result = provider._parse_response(
        {
            "id": "msg_ws",
            "content": [
                {
                    "type": "server_tool_use",
                    "id": "srvtoolu_1",
                    "name": "web_search_20250305",
                    "input": {"query": "test query"},
                },
                {
                    "type": "web_search_tool_result",
                    "tool_use_id": "srvtoolu_1",
                    "content": [
                        {
                            "type": "web_search_result",
                            "title": "Example Page",
                            "url": "https://example.com/page",
                            "page_snippet": "A snippet from the page.",
                        },
                        {
                            "type": "web_search_result",
                            "title": "Another Page",
                            "url": "https://example.com/another",
                            "page_snippet": "Another snippet.",
                        },
                    ],
                },
                {"type": "text", "text": "Here is the answer."},
            ],
            "usage": {
                "input_tokens": 100,
                "output_tokens": 30,
            },
        }
    )

    assert result.text == "Here is the answer."
    assert len(result.web_search_sources) == 2
    assert result.web_search_sources[0].title == "Example Page"
    assert result.web_search_sources[0].url == "https://example.com/page"
    assert result.web_search_sources[0].snippet == "A snippet from the page."
    assert result.web_search_sources[1].title == "Another Page"
    assert result.provider_data["display_items"] == [
        {
            "type": "web_search",
            "queries": ["test query"],
            "sources": [
                {
                    "title": "Example Page",
                    "url": "https://example.com/page",
                    "snippet": "A snippet from the page.",
                },
                {
                    "title": "Another Page",
                    "url": "https://example.com/another",
                    "snippet": "Another snippet.",
                },
            ],
        },
        {"type": "text", "text": "Here is the answer."},
    ]


def test_anthropic_web_search_sources_preserve_duplicate_urls() -> None:
    provider = AnthropicProvider(_make_settings())

    result = provider._parse_response(
        {
            "id": "msg_ws_dup",
            "content": [
                {
                    "type": "web_search_tool_result",
                    "tool_use_id": "srvtoolu_1",
                    "content": [
                        {
                            "type": "web_search_result",
                            "title": "Same Page",
                            "url": "https://example.com/same",
                            "page_snippet": "First snippet.",
                        },
                        {
                            "type": "web_search_result",
                            "title": "Same Page Again",
                            "url": "https://example.com/same",
                            "page_snippet": "Second snippet.",
                        },
                    ],
                },
                {"type": "text", "text": "Answer."},
            ],
            "usage": {
                "input_tokens": 10,
                "output_tokens": 5,
            },
        }
    )

    assert len(result.web_search_sources) == 2
    assert result.web_search_sources[0].url == "https://example.com/same"
    assert result.web_search_sources[1].url == "https://example.com/same"


def test_anthropic_server_tool_use_not_in_function_calls() -> None:
    provider = AnthropicProvider(_make_settings())

    result = provider._parse_response(
        {
            "id": "msg_ws2",
            "content": [
                {
                    "type": "server_tool_use",
                    "id": "srvtoolu_1",
                    "name": "web_search_20250305",
                    "input": {"query": "test"},
                },
                {
                    "type": "web_search_tool_result",
                    "tool_use_id": "srvtoolu_1",
                    "content": [],
                },
                {"type": "text", "text": "Done."},
            ],
            "usage": {
                "input_tokens": 10,
                "output_tokens": 5,
            },
        }
    )

    assert result.function_calls == []


def test_anthropic_request_turn_preserves_web_search_order_from_content_blocks(
    monkeypatch,
    display_spy,
    make_http_response,
) -> None:
    events: list[tuple[str, object]] = []

    original_render_markdown = display_spy.render_markdown
    original_function_result = display_spy.function_result

    def capture_markdown(text: str) -> None:
        events.append(("message", text))
        original_render_markdown(text)

    def capture_function_result(
        *,
        name: str,
        success: bool,
        call_id: str,
        arguments: object,
    ) -> None:
        events.append(("tool", name))
        original_function_result(
            name=name,
            success=success,
            call_id=call_id,
            arguments=arguments,
        )

    display_spy.render_markdown = capture_markdown
    display_spy.function_result = capture_function_result

    def fake_urlopen(
        request: urllib.request.Request,
        timeout: float,
    ):
        del request, timeout
        return make_http_response(
            {
                "id": "msg_ws_order",
                "content": [
                    {
                        "type": "text",
                        "text": "I'll search for the current Bitcoin price.",
                    },
                    {
                        "type": "server_tool_use",
                        "id": "srvtoolu_1",
                        "name": "web_search",
                        "input": {"query": "bitcoin price"},
                    },
                    {
                        "type": "web_search_tool_result",
                        "tool_use_id": "srvtoolu_1",
                        "content": [
                            {
                                "type": "web_search_result",
                                "title": "BTC",
                                "url": "https://example.com/btc",
                                "page_snippet": "Bitcoin price now.",
                            }
                        ],
                    },
                    {"type": "text", "text": "Bitcoin is $1."},
                ],
                "usage": {"input_tokens": 8, "output_tokens": 3},
            }
        )

    monkeypatch.setattr(urllib.request, "urlopen", fake_urlopen)

    provider = AnthropicProvider(_make_settings(model="claude-haiku-4-5"))
    provider.request_turn(
        user_message="check bitcoin price",
        display=display_spy,
        session_usage=TokenUsage(),
        turn_usage=TokenUsage(),
    )

    assert events == [
        ("message", "I'll search for the current Bitcoin price."),
        ("tool", "web_search"),
        ("message", "Bitcoin is $1."),
    ]



from __future__ import annotations

import subprocess
from pathlib import Path

import pytest

from pbi_agent.tools import shell as shell_tool
from pbi_agent.tools.types import ToolContext


def test_shell_handle_runs_command_with_workspace_defaults(
    tmp_path: Path,
    monkeypatch,
) -> None:
    monkeypatch.chdir(tmp_path)
    calls: list[dict[str, object]] = []

    def fake_run(
        command: str,
        *,
        cwd: str,
        capture_output: bool,
        text: bool,
        shell: bool,
        timeout: float,
    ) -> subprocess.CompletedProcess[bytes]:
        calls.append(
            {
                "command": command,
                "cwd": cwd,
                "capture_output": capture_output,
                "text": text,
                "shell": shell,
                "timeout": timeout,
            }
        )
        return subprocess.CompletedProcess(
            args=command,
            returncode=0,
            stdout=b"stdout text",
            stderr=b"stderr text",
        )

    monkeypatch.setattr(shell_tool.subprocess, "run", fake_run)

    result = shell_tool.handle({"command": "echo hi"}, ToolContext())

    assert result == {
        "stdout": "stdout text",
        "stderr": "stderr text",
        "exit_code": 0,
    }
    assert calls == [
        {
            "command": "echo hi",
            "cwd": str(tmp_path.resolve()),
            "capture_output": True,
            "text": False,
            "shell": True,
            "timeout": shell_tool.MAX_TIMEOUT_MS / 1000.0,
        }
    ]


def test_shell_handle_uses_requested_directory_and_clamps_timeout(
    tmp_path: Path,
    monkeypatch,
) -> None:
    monkeypatch.chdir(tmp_path)
    workdir = tmp_path / "nested"
    workdir.mkdir()
    seen: dict[str, object] = {}

    def fake_run(
        command: str,
        *,
        cwd: str,
        capture_output: bool,
        text: bool,
        shell: bool,
        timeout: float,
    ) -> subprocess.CompletedProcess[bytes]:
        del capture_output, text, shell
        seen["command"] = command
        seen["cwd"] = cwd
        seen["timeout"] = timeout
        return subprocess.CompletedProcess(
            args=command,
            returncode=0,
            stdout=b"",
            stderr=b"",
        )

    monkeypatch.setattr(shell_tool.subprocess, "run", fake_run)

    shell_tool.handle(
        {
            "command": "pwd",
            "working_directory": "nested",
            "timeout_ms": shell_tool.MAX_TIMEOUT_MS * 2,
        },
        ToolContext(),
    )

    assert seen == {
        "command": "pwd",
        "cwd": str(workdir.resolve()),
        "timeout": shell_tool.MAX_TIMEOUT_MS / 1000.0,
    }


def test_shell_handle_returns_timeout_payload(tmp_path: Path, monkeypatch) -> None:
    monkeypatch.chdir(tmp_path)

    def fake_run(*args: object, **kwargs: object) -> subprocess.CompletedProcess[bytes]:
        del args, kwargs
        raise subprocess.TimeoutExpired(
            cmd="sleep 10",
            timeout=0.5,
            output=b"partial stdout",
            stderr=b"partial stderr",
        )

    monkeypatch.setattr(shell_tool.subprocess, "run", fake_run)

    result = shell_tool.handle(
        {"command": "sleep 10", "timeout_ms": 500},
        ToolContext(),
    )

    assert result == {
        "stdout": "partial stdout",
        "stderr": "partial stderr",
        "exit_code": None,
        "timed_out": True,
        "error": "Command timed out after 500ms.",
    }


def test_shell_handle_returns_error_payload_for_subprocess_failure(
    tmp_path: Path,
    monkeypatch,
) -> None:
    monkeypatch.chdir(tmp_path)

    def fake_run(*args: object, **kwargs: object) -> subprocess.CompletedProcess[bytes]:
        del args, kwargs
        raise OSError("boom")

    monkeypatch.setattr(shell_tool.subprocess, "run", fake_run)

    result = shell_tool.handle({"command": "bad-command"}, ToolContext())

    assert result["stdout"] == ""
    assert result["stderr"] == "boom"
    assert result["exit_code"] == 1
    assert result["error"] == "Shell execution failed: boom"


def test_shell_handle_bounds_large_stdout_and_stderr(
    tmp_path: Path,
    monkeypatch,
) -> None:
    monkeypatch.chdir(tmp_path)
    stdout = f"stdout-start-{'a' * (shell_tool.MAX_OUTPUT_CHARS + 50)}-stdout-end"
    stderr = f"stderr-start-{'b' * (shell_tool.MAX_OUTPUT_CHARS + 50)}-stderr-end"

    def fake_run(*args: object, **kwargs: object) -> subprocess.CompletedProcess[bytes]:
        del args, kwargs
        return subprocess.CompletedProcess(
            args="python",
            returncode=0,
            stdout=stdout.encode("utf-8"),
            stderr=stderr.encode("utf-8"),
        )

    monkeypatch.setattr(shell_tool.subprocess, "run", fake_run)

    result = shell_tool.handle({"command": "python"}, ToolContext())

    assert result["exit_code"] == 0
    assert result["stdout_truncated"] is True
    assert result["stderr_truncated"] is True
    assert len(result["stdout"]) <= shell_tool.MAX_OUTPUT_CHARS
    assert len(result["stderr"]) <= shell_tool.MAX_OUTPUT_CHARS
    assert result["stdout"].startswith("stdout-start-")
    assert result["stdout"].endswith("-stdout-end")
    assert result["stderr"].startswith("stderr-start-")
    assert result["stderr"].endswith("-stderr-end")
    assert "chars omitted" in result["stdout"]
    assert "chars omitted" in result["stderr"]


def test_resolve_working_directory_rejects_workspace_escape(tmp_path: Path) -> None:
    outside = tmp_path.parent

    with pytest.raises(ValueError, match="outside workspace"):
        shell_tool._resolve_working_directory(tmp_path.resolve(), str(outside))



from __future__ import annotations

from pbi_agent.tools import registry


def test_registry_exposes_expected_built_in_tools() -> None:
    expected = {
        "skill_knowledge",
        "init_report",
        "shell",
        "python_exec",
        "apply_patch",
        "list_files",
        "search_files",
        "read_file",
        "read_web_url",
        "sub_agent",
    }

    assert expected.issubset({spec.name for spec in registry.get_tool_specs()})
    assert expected.issubset(
        {item["name"] for item in registry.get_openai_tool_definitions()}
    )
    assert expected.issubset(
        {item["name"] for item in registry.get_anthropic_tool_definitions()}
    )
    assert expected.issubset(
        {
            item["function"]["name"]
            for item in registry.get_openai_chat_tool_definitions()
        }
    )


def test_registry_returns_none_for_unknown_tool() -> None:
    assert registry.get_tool_handler("missing_tool") is None
    assert registry.get_tool_spec("missing_tool") is None




from __future__ import annotations

from pathlib import Path


from pbi_agent.tools import apply_patch as apply_patch_tool
from pbi_agent.tools.output import MAX_OUTPUT_CHARS
from pbi_agent.tools.types import ToolContext


def test_apply_patch_handle_create_update_and_delete_file(
    tmp_path: Path,
    monkeypatch,
) -> None:
    monkeypatch.chdir(tmp_path)

    create_result = apply_patch_tool.handle(
        {
            "operation_type": "create_file",
            "path": "notes/example.txt",
            "diff": "+hello\n+world",
        },
        ToolContext(),
    )

    created_file = tmp_path / "notes" / "example.txt"
    assert create_result["status"] == "completed"
    assert created_file.read_text(encoding="utf-8") == "hello\nworld"

    update_result = apply_patch_tool.handle(
        {
            "operation_type": "update_file",
            "path": "notes/example.txt",
            "diff": " hello\n-world\n+there",
        },
        ToolContext(),
    )

    assert update_result["status"] == "completed"
    assert created_file.read_text(encoding="utf-8") == "hello\nthere"

    delete_result = apply_patch_tool.handle(
        {
            "operation_type": "delete_file",
            "path": "notes/example.txt",
        },
        ToolContext(),
    )

    assert delete_result["status"] == "completed"
    assert not created_file.exists()


def test_apply_patch_handle_rejects_paths_outside_workspace(
    tmp_path: Path,
    monkeypatch,
) -> None:
    monkeypatch.chdir(tmp_path)
    outside_path = tmp_path.parent / "escape.txt"

    result = apply_patch_tool.handle(
        {
            "operation_type": "create_file",
            "path": str(outside_path),
            "diff": "+secret",
        },
        ToolContext(),
    )

    assert result["status"] == "failed"
    assert "outside workspace" in result["error"]


def test_apply_patch_handle_validates_required_arguments() -> None:
    missing_path = apply_patch_tool.handle(
        {
            "operation_type": "create_file",
            "path": "",
            "diff": "+hello",
        },
        ToolContext(),
    )
    unsupported_operation = apply_patch_tool.handle(
        {
            "operation_type": "rename_file",
            "path": "example.txt",
        },
        ToolContext(),
    )

    assert missing_path == {"error": "'path' must be a non-empty string."}
    assert unsupported_operation == {
        "error": "Unsupported operation_type 'rename_file'."
    }


def test_apply_patch_handle_bounds_long_error_output(monkeypatch) -> None:
    def raise_long_error(path: Path, diff: str | None) -> None:
        raise ValueError(f"start-{'x' * (MAX_OUTPUT_CHARS + 200)}-end")

    monkeypatch.setattr(apply_patch_tool, "_create_file", raise_long_error)

    result = apply_patch_tool.handle(
        {
            "operation_type": "create_file",
            "path": "notes/example.txt",
            "diff": "+hello",
        },
        ToolContext(),
    )

    assert result["status"] == "failed"
    assert len(result["error"]) <= MAX_OUTPUT_CHARS
    assert result["error"].startswith("start-")
    assert result["error"].endswith("-end")
    assert "chars omitted" in result["error"]



from __future__ import annotations

import importlib.resources
import json
import threading
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any

_ZERO_PRICING: tuple[float, float, float, float, float] = (0.0, 0.0, 0.0, 0.0, 0.0)


class ModelCatalog:
    """Registry of model pricing and context window sizes.

    Loads a bundled ``model_catalog.json`` and optionally merges user
    overrides from ``~/.pbi-agent/model_catalog.json``.
    """

    def __init__(self) -> None:
        self._models: dict[str, dict] = {}
        self._default_context_window: int = 200_000
        self._load_bundled()
        self._load_user_overrides()

    # ------------------------------------------------------------------
    def _load_bundled(self) -> None:
        ref = importlib.resources.files("pbi_agent.models").joinpath(
            "model_catalog.json"
        )
        data = json.loads(ref.read_text(encoding="utf-8"))
        self._models.update(data.get("models", {}))
        self._default_context_window = data.get("defaults", {}).get(
            "context_window", self._default_context_window
        )

    def _load_user_overrides(self) -> None:
        user_path = Path.home() / ".pbi-agent" / "model_catalog.json"
        if not user_path.is_file():
            return
        try:
            data = json.loads(user_path.read_text(encoding="utf-8"))
            self._models.update(data.get("models", {}))
        except (json.JSONDecodeError, OSError):
            pass

    # ------------------------------------------------------------------
    def _find_entry(self, model: str) -> dict | None:
        if model in self._models:
            return self._models[model]
        for key, entry in self._models.items():
            if model.startswith(key):
                return entry
        return None

    def get_pricing(self, model: str) -> tuple[float, float, float, float, float]:
        """Return per-MTok prices ``(input, cache_write_5m, cache_write_1h,
        cache_hit, output)`` for *model*.

        Returns all zeros for unknown models.
        """
        entry = self._find_entry(model)
        if entry is None:
            return _ZERO_PRICING
        p = entry.get("pricing")
        if p is None:
            return _ZERO_PRICING
        return (
            p.get("input", 0.0),
            p.get("cache_write_5m", 0.0),
            p.get("cache_write_1h", 0.0),
            p.get("cache_hit", 0.0),
            p.get("output", 0.0),
        )

    def get_context_window(self, model: str) -> int:
        """Return the context window size for *model*."""
        entry = self._find_entry(model)
        if entry is not None:
            return entry.get("context_window", self._default_context_window)
        return self._default_context_window


# Lazy singleton — avoids file I/O at import time.
_catalog: ModelCatalog | None = None


def _get_catalog() -> ModelCatalog:
    global _catalog
    if _catalog is None:
        _catalog = ModelCatalog()
    return _catalog


def context_window_for_model(model: str) -> int:
    """Return the context window size for *model*."""
    return _get_catalog().get_context_window(model)


def _pricing_for_model(model: str) -> tuple[float, float, float, float, float]:
    """Return per-MTok prices for *model*.

    Returns ``(base_input, cache_write_5m, cache_write_1h, cache_hit, output)``.
    """
    return _get_catalog().get_pricing(model)


def _service_tier_multiplier(service_tier: str) -> float:
    """Return the pricing multiplier for an OpenAI service tier."""
    if service_tier == "flex":
        return 0.5
    if service_tier == "priority":
        return 2.0
    return 1.0


def _estimated_cost(
    *,
    model: str,
    input_tokens: int,
    cached_input_tokens: int,
    cache_write_tokens: int,
    cache_write_1h_tokens: int,
    output_tokens: int,
    service_tier: str = "",
) -> float:
    p_in, p_w5, p_w1h, p_hit, p_out = _pricing_for_model(model)
    non_cached_input_tokens = max(
        input_tokens - cached_input_tokens - cache_write_tokens - cache_write_1h_tokens,
        0,
    )
    cost = (
        (non_cached_input_tokens / 1_000_000.0) * p_in
        + (cache_write_tokens / 1_000_000.0) * p_w5
        + (cache_write_1h_tokens / 1_000_000.0) * p_w1h
        + (cached_input_tokens / 1_000_000.0) * p_hit
        + (output_tokens / 1_000_000.0) * p_out
    )
    if service_tier:
        cost *= _service_tier_multiplier(service_tier)
    return cost


@dataclass(slots=True)
class TokenUsage:
    input_tokens: int = 0
    cached_input_tokens: int = 0
    cache_write_tokens: int = 0
    cache_write_1h_tokens: int = 0
    output_tokens: int = 0
    reasoning_tokens: int = 0
    tool_use_tokens: int = 0
    provider_total_tokens: int = 0
    sub_agent_input_tokens: int = 0
    sub_agent_cached_input_tokens: int = 0
    sub_agent_cache_write_tokens: int = 0
    sub_agent_cache_write_1h_tokens: int = 0
    sub_agent_output_tokens: int = 0
    sub_agent_reasoning_tokens: int = 0
    sub_agent_tool_use_tokens: int = 0
    sub_agent_provider_total_tokens: int = 0
    sub_agent_cost_usd: float = 0.0
    context_tokens: int = 0
    model: str = ""
    service_tier: str = ""
    _lock: threading.Lock = field(
        default_factory=threading.Lock,
        init=False,
        repr=False,
        compare=False,
    )

    @property
    def non_cached_input_tokens(self) -> int:
        """Base input tokens charged at the standard input rate.

        For Anthropic, ``input_tokens`` is set to the total input
        (base + cache reads + cache writes) so we subtract out all
        cached / cache-creation components.  For OpenAI the cache-write
        fields are always 0, so this collapses to the old behaviour.
        """
        return max(
            self.input_tokens
            - self.cached_input_tokens
            - self.cache_write_tokens
            - self.cache_write_1h_tokens,
            0,
        )

    @property
    def total_tokens(self) -> int:
        if self.provider_total_tokens > 0:
            return self.provider_total_tokens
        return self.input_tokens + self.output_tokens

    @property
    def estimated_cost_usd(self) -> float:
        main_input_tokens = max(self.input_tokens - self.sub_agent_input_tokens, 0)
        main_cached_input_tokens = max(
            self.cached_input_tokens - self.sub_agent_cached_input_tokens,
            0,
        )
        main_cache_write_tokens = max(
            self.cache_write_tokens - self.sub_agent_cache_write_tokens,
            0,
        )
        main_cache_write_1h_tokens = max(
            self.cache_write_1h_tokens - self.sub_agent_cache_write_1h_tokens,
            0,
        )
        main_output_tokens = max(self.output_tokens - self.sub_agent_output_tokens, 0)
        return (
            _estimated_cost(
                model=self.model,
                input_tokens=main_input_tokens,
                cached_input_tokens=main_cached_input_tokens,
                cache_write_tokens=main_cache_write_tokens,
                cache_write_1h_tokens=main_cache_write_1h_tokens,
                output_tokens=main_output_tokens,
                service_tier=self.service_tier,
            )
            + self.sub_agent_cost_usd
        )

    def add(self, other: "TokenUsage") -> None:
        self._add_snapshot(other.snapshot(), as_sub_agent=False)

    def add_sub_agent(self, other: "TokenUsage") -> None:
        self._add_snapshot(other.snapshot(), as_sub_agent=True)

    @property
    def sub_agent_total_tokens(self) -> int:
        if self.sub_agent_provider_total_tokens > 0:
            return self.sub_agent_provider_total_tokens
        return self.sub_agent_input_tokens + self.sub_agent_output_tokens

    @property
    def main_agent_total_tokens(self) -> int:
        return max(self.total_tokens - self.sub_agent_total_tokens, 0)

    def _add_snapshot(
        self, other_snapshot: "TokenUsage", *, as_sub_agent: bool
    ) -> None:
        with self._lock:
            self.input_tokens += other_snapshot.input_tokens
            self.cached_input_tokens += other_snapshot.cached_input_tokens
            self.cache_write_tokens += other_snapshot.cache_write_tokens
            self.cache_write_1h_tokens += other_snapshot.cache_write_1h_tokens
            self.output_tokens += other_snapshot.output_tokens
            self.reasoning_tokens += other_snapshot.reasoning_tokens
            self.tool_use_tokens += other_snapshot.tool_use_tokens
            self.provider_total_tokens += other_snapshot.provider_total_tokens
            if as_sub_agent:
                self.sub_agent_input_tokens += other_snapshot.input_tokens
                self.sub_agent_cached_input_tokens += other_snapshot.cached_input_tokens
                self.sub_agent_cache_write_tokens += other_snapshot.cache_write_tokens
                self.sub_agent_cache_write_1h_tokens += (
                    other_snapshot.cache_write_1h_tokens
                )
                self.sub_agent_output_tokens += other_snapshot.output_tokens
                self.sub_agent_reasoning_tokens += other_snapshot.reasoning_tokens
                self.sub_agent_tool_use_tokens += other_snapshot.tool_use_tokens
                self.sub_agent_provider_total_tokens += (
                    other_snapshot.provider_total_tokens
                )
                self.sub_agent_cost_usd += other_snapshot.estimated_cost_usd
            else:
                self.sub_agent_cost_usd += other_snapshot.sub_agent_cost_usd
            if other_snapshot.context_tokens and not as_sub_agent:
                self.context_tokens = other_snapshot.context_tokens
            if not self.model and other_snapshot.model:
                self.model = other_snapshot.model
            if not self.service_tier and other_snapshot.service_tier:
                self.service_tier = other_snapshot.service_tier

    def snapshot(self) -> "TokenUsage":
        with self._lock:
            return TokenUsage(
                input_tokens=self.input_tokens,
                cached_input_tokens=self.cached_input_tokens,
                cache_write_tokens=self.cache_write_tokens,
                cache_write_1h_tokens=self.cache_write_1h_tokens,
                output_tokens=self.output_tokens,
                reasoning_tokens=self.reasoning_tokens,
                tool_use_tokens=self.tool_use_tokens,
                provider_total_tokens=self.provider_total_tokens,
                sub_agent_input_tokens=self.sub_agent_input_tokens,
                sub_agent_cached_input_tokens=self.sub_agent_cached_input_tokens,
                sub_agent_cache_write_tokens=self.sub_agent_cache_write_tokens,
                sub_agent_cache_write_1h_tokens=self.sub_agent_cache_write_1h_tokens,
                sub_agent_output_tokens=self.sub_agent_output_tokens,
                sub_agent_reasoning_tokens=self.sub_agent_reasoning_tokens,
                sub_agent_tool_use_tokens=self.sub_agent_tool_use_tokens,
                sub_agent_provider_total_tokens=self.sub_agent_provider_total_tokens,
                sub_agent_cost_usd=self.sub_agent_cost_usd,
                context_tokens=self.context_tokens,
                model=self.model,
                service_tier=self.service_tier,
            )


@dataclass(slots=True)
class ToolCall:
    call_id: str
    name: str
    arguments: dict | str | None


@dataclass(slots=True)
class ImageAttachment:
    path: str
    mime_type: str
    data_base64: str
    byte_count: int = 0


@dataclass(slots=True)
class UserTurnInput:
    text: str = ""
    images: list[ImageAttachment] = field(default_factory=list)


@dataclass(slots=True)
class WebSearchSource:
    """A single citation/source returned by a provider's native web search."""

    title: str
    url: str
    snippet: str = ""


@dataclass(slots=True)
class CompletedResponse:
    response_id: str | None
    text: str
    assistant_messages: list[str] = field(default_factory=list)
    usage: TokenUsage = field(default_factory=TokenUsage)
    function_calls: list[ToolCall] = field(default_factory=list)
    # Reasoning summary text extracted from OpenAI ``reasoning`` output items.
    reasoning_summary: str = ""
    # Detailed reasoning text extracted from OpenAI ``reasoning.content`` items.
    reasoning_content: str = ""
    # Provider-specific opaque data (e.g. raw Anthropic content blocks for
    # history replay).  The session layer never inspects this; only the
    # provider that created the response uses it.
    provider_data: Any = None
    # Web search citations returned by the provider's native search tool.
    web_search_sources: list[WebSearchSource] = field(default_factory=list)
    # Whether the provider response included a hosted web search call item.
    had_web_search_call: bool = False

    @property
    def has_tool_calls(self) -> bool:
        return bool(self.function_calls)


@dataclass(slots=True)
class AgentOutcome:
    response_id: str | None
    text: str
    tool_errors: bool = False



"""xAI Responses HTTP provider.

Uses direct synchronous HTTP calls to xAI's Responses API. Conversation
history is managed server-side via ``previous_response_id``.
"""

from __future__ import annotations

import json
import re
import time
import urllib.error
import urllib.parse
import urllib.request
from typing import Any

from pbi_agent import __version__
from pbi_agent.agent.system_prompt import get_system_prompt
from pbi_agent.agent.tool_runtime import (
    execute_tool_calls as _execute_tool_calls,
    to_function_call_output_items,
)
from pbi_agent.config import Settings
from pbi_agent.models.messages import (
    CompletedResponse,
    TokenUsage,
    ToolCall,
    UserTurnInput,
    WebSearchSource,
)
from pbi_agent.providers.base import Provider
from pbi_agent.tools.registry import get_openai_tool_definitions
from pbi_agent.tools.types import ToolContext
from pbi_agent.ui.display_protocol import DisplayProtocol

_REQUEST_TIMEOUT_SECS = 3600.0
_REASONING_EFFORT_MODELS = ("grok-3-mini",)
_ENCRYPTED_REASONING_MODELS = ("grok-4",)
_EFFORT_MAP: dict[str, str] = {
    "low": "low",
    "medium": "high",
    "high": "high",
    "xhigh": "high",
}
_HTTP_ERROR_TYPES: dict[int, str] = {
    400: "invalid_request_error",
    401: "authentication_error",
    403: "permission_error",
    404: "not_found_error",
    405: "invalid_request_error",
    415: "invalid_request_error",
    422: "invalid_request_error",
    429: "rate_limit_error",
}
_HTTP_ERROR_MESSAGES: dict[int, str] = {
    400: "Request body or URL contains invalid input.",
    401: "No authorization header or an invalid authorization token was provided.",
    403: "API key or team lacks permission to perform this action.",
    404: "The requested model or endpoint could not be found.",
    405: "HTTP method is not allowed for this endpoint.",
    415: "Request body is missing or Content-Type is not application/json.",
    422: "Request body contains an invalid field format.",
    429: "Too many requests. Reduce request rate or increase your rate limit.",
}


class XAIProvider(Provider):
    """Provider backed by xAI's synchronous Responses HTTP API."""

    def __init__(
        self,
        settings: Settings,
        *,
        system_prompt: str | None = None,
        excluded_tools: set[str] | None = None,
    ) -> None:
        self._settings = settings
        self._tools = get_openai_tool_definitions(excluded_names=excluded_tools)
        if settings.web_search:
            self._tools.append({"type": "web_search"})
        self._instructions = system_prompt or get_system_prompt()
        self._previous_response_id: str | None = None

    @property
    def settings(self) -> Settings:
        return self._settings

    def set_previous_response_id(self, response_id: str | None) -> None:
        self._previous_response_id = response_id

    def connect(self) -> None:
        if not self._settings.api_key:
            raise ValueError(
                "Missing API key. Set PBI_AGENT_API_KEY in environment or pass "
                "--api-key."
            )

    def close(self) -> None:
        pass

    def reset_conversation(self) -> None:
        self._previous_response_id = None

    def request_turn(
        self,
        *,
        user_message: str | None = None,
        user_input: UserTurnInput | None = None,
        tool_result_items: list[dict[str, Any]] | None = None,
        instructions: str | None = None,
        display: DisplayProtocol,
        session_usage: TokenUsage,
        turn_usage: TokenUsage,
    ) -> CompletedResponse:
        if user_input is None and user_message is not None:
            user_input = UserTurnInput(text=user_message)

        if user_input is not None:
            if user_input.images:
                raise ValueError("xAI image inputs are not enabled in this build.")
            input_items = [_build_user_input_item(user_input.text)]
        elif tool_result_items is not None:
            input_items = tool_result_items
        else:
            raise ValueError("Either user_input or tool_result_items is required")

        result = self._http_request(
            input_items=input_items,
            instructions=instructions or self._instructions,
            display=display,
        )
        self._previous_response_id = result.response_id
        session_usage.add(result.usage)
        turn_usage.add(result.usage)
        display.session_usage(session_usage)

        if result.reasoning_summary or result.reasoning_content:
            display.render_thinking(
                _reasoning_body_text(
                    result.reasoning_content,
                    result.reasoning_summary,
                ),
                title=result.reasoning_summary or None,
            )

        display_items = result.provider_data.get("display_items")
        if isinstance(display_items, list) and display_items:
            for item in display_items:
                if not isinstance(item, dict):
                    continue
                item_type = item.get("type")
                if item_type == "message":
                    text = item.get("text")
                    if isinstance(text, str) and text:
                        display.render_markdown(text)
                elif item_type == "web_search_call":
                    _display_web_search_result(
                        display,
                        item.get("sources", []),
                        queries=item.get("queries", []),
                    )
        else:
            if result.text:
                display.render_markdown(result.text)

            if result.had_web_search_call or result.web_search_sources:
                _display_web_search_result(display, result.web_search_sources)

        return result

    def execute_tool_calls(
        self,
        response: CompletedResponse,
        *,
        max_workers: int,
        display: DisplayProtocol,
        session_usage: TokenUsage,
        turn_usage: TokenUsage,
        sub_agent_depth: int = 0,
    ) -> tuple[list[dict[str, Any]], bool]:
        if not response.function_calls:
            return [], False

        displayable_calls = [
            call for call in response.function_calls if call.name != "sub_agent"
        ]
        if displayable_calls:
            display.function_start(len(displayable_calls))
        batch = _execute_tool_calls(
            response.function_calls,
            max_workers=max_workers,
            context=ToolContext(
                settings=self._settings,
                display=display,
                session_usage=session_usage,
                turn_usage=turn_usage,
                sub_agent_depth=sub_agent_depth,
            ),
        )

        for result in batch.results:
            call = _find_by_id(response.function_calls, result.call_id)
            if call and call.name == "sub_agent":
                continue
            display.function_result(
                name=call.name if call else "unknown",
                success=not result.is_error,
                call_id=result.call_id,
                arguments=call.arguments if call else None,
            )
        if displayable_calls:
            display.tool_group_end()

        return to_function_call_output_items(batch.results), batch.had_errors

    def _http_request(
        self,
        *,
        input_items: list[dict[str, Any]],
        instructions: str,
        display: DisplayProtocol,
    ) -> CompletedResponse:
        display.wait_start(_waiting_message_for_input_items(input_items))

        body = self._build_request_body(
            input_items=input_items,
            instructions=instructions,
        )
        request_data = json.dumps(body).encode("utf-8")
        headers = {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self._settings.api_key}",
            "User-Agent": f"pbi-agent/{__version__}",
        }

        max_retries = self._settings.max_retries
        last_error: Exception | None = None
        last_error_message: str | None = None

        for attempt in range(max_retries + 1):
            if attempt > 0:
                display.retry_notice(attempt, max_retries)

            try:
                req = urllib.request.Request(
                    self._settings.responses_url,
                    data=request_data,
                    headers=headers,
                    method="POST",
                )
                with urllib.request.urlopen(req, timeout=_REQUEST_TIMEOUT_SECS) as resp:
                    response_json = json.loads(resp.read().decode("utf-8"))

                result = self._parse_response(response_json)
                display.wait_stop()
                return result
            except urllib.error.HTTPError as exc:
                error_body = _read_error_body(exc)
                error_payload = _normalize_http_error(exc, error_body)
                if exc.code == 429:
                    if attempt >= max_retries:
                        display.wait_stop()
                        raise RuntimeError(
                            _format_error_message(
                                f"xAI rate limit exceeded after {max_retries + 1} attempts",
                                error_payload,
                            )
                        ) from exc
                    wait = _extract_retry_after(exc, attempt)
                    display.rate_limit_notice(
                        wait_seconds=wait,
                        attempt=attempt + 1,
                        max_retries=max_retries,
                    )
                    time.sleep(wait)
                    continue

                if exc.code >= 500:
                    last_error = exc
                    last_error_message = _format_error_message(
                        f"xAI request failed after {max_retries + 1} attempts",
                        error_payload,
                    )
                    continue

                display.wait_stop()
                raise RuntimeError(
                    _format_error_message(
                        f"xAI Responses API error {exc.code}",
                        error_payload,
                    )
                ) from exc
            except urllib.error.URLError as exc:
                last_error = exc
                last_error_message = None
                continue

        display.wait_stop()
        if last_error is not None:
            if last_error_message:
                raise RuntimeError(last_error_message) from last_error
            raise RuntimeError(
                f"xAI request failed after {max_retries + 1} attempts: {last_error}"
            ) from last_error
        raise RuntimeError("xAI request failed after retries.")

    def _build_request_body(
        self,
        *,
        input_items: list[dict[str, Any]],
        instructions: str | None,
    ) -> dict[str, Any]:
        request_input_items = list(input_items)
        if instructions and not self._previous_response_id:
            request_input_items.insert(0, _build_system_input_item(instructions))

        body: dict[str, Any] = {
            "model": self._settings.model,
            "max_output_tokens": self._settings.max_tokens,
            "input": request_input_items,
            "tools": self._tools,
            "parallel_tool_calls": True,
            "stream": False,
        }
        if self._previous_response_id:
            body["previous_response_id"] = self._previous_response_id

        include = _response_include(self._settings.model)
        if include:
            body["include"] = include

        reasoning = _reasoning_request(
            self._settings.model,
            self._settings.reasoning_effort,
        )
        if reasoning:
            body["reasoning"] = reasoning

        return body

    def _parse_response(self, response_json: dict[str, Any]) -> CompletedResponse:
        text_parts: list[str] = []
        reasoning_summary_parts: list[str] = []
        reasoning_content_parts: list[str] = []
        encrypted_reasoning_parts: list[str] = []
        function_calls: list[ToolCall] = []
        web_search_sources: list[WebSearchSource] = []
        had_web_search_call = False
        display_items: list[dict[str, Any]] = []
        last_web_search_index: int | None = None

        output_items = response_json.get("output", [])
        if not isinstance(output_items, list):
            output_items = []

        for item in output_items:
            if not isinstance(item, dict):
                continue
            item_type = item.get("type")

            if item_type == "reasoning":
                reasoning_summary_parts.extend(
                    _extract_reasoning_summary_texts(item.get("summary"))
                )
                encrypted_content = item.get("encrypted_content")
                if isinstance(encrypted_content, str) and encrypted_content:
                    encrypted_reasoning_parts.append(encrypted_content)

                for content_entry in item.get("content", []):
                    if not isinstance(content_entry, dict):
                        continue
                    if content_entry.get("type") == "reasoning_text":
                        reasoning_text = content_entry.get("text", "")
                        if reasoning_text:
                            reasoning_content_parts.append(reasoning_text)

            elif item_type == "message":
                message_text_parts: list[str] = []
                message_annotation_sources: list[WebSearchSource] = []
                for part in item.get("content", []):
                    if not isinstance(part, dict):
                        continue
                    if part.get("type") == "output_text":
                        text = part.get("text", "")
                        if text:
                            text_parts.append(text)
                            message_text_parts.append(text)
                        for annotation in part.get("annotations", []):
                            if not isinstance(annotation, dict):
                                continue
                            if annotation.get("type") == "url_citation":
                                source = _annotation_web_search_source(annotation)
                                if source is not None:
                                    message_annotation_sources.append(source)
                if message_annotation_sources:
                    if last_web_search_index is None:
                        display_items.append(
                            {
                                "type": "web_search_call",
                                "queries": [],
                                "sources": _serialize_web_search_sources(
                                    message_annotation_sources
                                ),
                            }
                        )
                        last_web_search_index = len(display_items) - 1
                    else:
                        existing_sources = _deserialize_web_search_sources(
                            display_items[last_web_search_index].get("sources", [])
                        )
                        display_items[last_web_search_index]["sources"] = (
                            _serialize_web_search_sources(
                                _merge_web_search_sources(
                                    existing_sources,
                                    message_annotation_sources,
                                )
                            )
                        )
                    web_search_sources = _merge_web_search_sources(
                        web_search_sources,
                        message_annotation_sources,
                    )
                message_text = "".join(message_text_parts).strip()
                if message_text:
                    display_items.append({"type": "message", "text": message_text})

            elif item_type == "function_call":
                function_calls.append(_parse_function_call(item))

            elif item_type == "web_search_call":
                had_web_search_call = True
                item_sources = _extract_web_search_sources(item)
                item_queries = _extract_web_search_queries(item)
                web_search_sources = _merge_web_search_sources(
                    web_search_sources,
                    item_sources,
                )
                if (
                    last_web_search_index is not None
                    and last_web_search_index == len(display_items) - 1
                    and display_items[last_web_search_index].get("type")
                    == "web_search_call"
                    and display_items[last_web_search_index].get("queries", [])
                    == item_queries
                ):
                    existing_sources = _deserialize_web_search_sources(
                        display_items[last_web_search_index].get("sources", [])
                    )
                    display_items[last_web_search_index]["sources"] = (
                        _serialize_web_search_sources(
                            _merge_web_search_sources(existing_sources, item_sources)
                        )
                    )
                else:
                    display_items.append(
                        {
                            "type": "web_search_call",
                            "queries": item_queries,
                            "sources": _serialize_web_search_sources(item_sources),
                        }
                    )
                    last_web_search_index = len(display_items) - 1

        usage_obj = response_json.get("usage", {})
        input_tokens = int(_usage_value(usage_obj, "input_tokens"))
        output_tokens = int(_usage_value(usage_obj, "output_tokens"))
        total_tokens = int(_usage_value(usage_obj, "total_tokens"))
        input_details = usage_obj.get("input_tokens_details", {})
        output_details = usage_obj.get("output_tokens_details", {})

        cached_input_tokens = (
            int(input_details.get("cached_tokens", 0) or 0)
            if isinstance(input_details, dict)
            else 0
        )
        reasoning_tokens = (
            int(output_details.get("reasoning_tokens", 0) or 0)
            if isinstance(output_details, dict)
            else 0
        )

        reasoning_summary = "\n\n".join(
            part for part in reasoning_summary_parts if part.strip()
        ).strip()
        reasoning_content = "\n\n".join(
            part for part in reasoning_content_parts if part.strip()
        ).strip()
        text = "".join(text_parts).strip()
        if not text:
            output_text = response_json.get("output_text")
            if isinstance(output_text, str):
                text = output_text.strip()

        return CompletedResponse(
            response_id=response_json.get("id"),
            text=text,
            usage=TokenUsage(
                input_tokens=input_tokens,
                cached_input_tokens=cached_input_tokens,
                output_tokens=output_tokens,
                reasoning_tokens=reasoning_tokens,
                context_tokens=total_tokens or (input_tokens + output_tokens),
                model=_response_model_name(response_json),
            ),
            function_calls=function_calls,
            reasoning_summary=reasoning_summary,
            reasoning_content=reasoning_content,
            provider_data={
                "encrypted_reasoning_content": encrypted_reasoning_parts,
                "reasoning": response_json.get("reasoning"),
                "display_items": display_items,
            },
            web_search_sources=web_search_sources,
            had_web_search_call=had_web_search_call,
        )


def _build_user_input_item(prompt: str) -> dict[str, Any]:
    return {"role": "user", "content": prompt}


def _build_system_input_item(prompt: str) -> dict[str, Any]:
    return {"role": "system", "content": prompt}


def _extract_reasoning_summary_texts(raw_summary: Any) -> list[str]:
    if not isinstance(raw_summary, list):
        return []

    summary_parts: list[str] = []
    for entry in raw_summary:
        if isinstance(entry, dict):
            if entry.get("type") == "summary_text":
                text = entry.get("text", "")
                if text:
                    summary_parts.append(text)
        elif isinstance(entry, str) and entry:
            summary_parts.append(entry)
    return summary_parts


def _parse_function_call(item: dict[str, Any]) -> ToolCall:
    raw_args = item.get("arguments", "")
    arguments: dict[str, Any] | str | None
    if isinstance(raw_args, str):
        try:
            arguments = json.loads(raw_args) if raw_args else {}
        except json.JSONDecodeError:
            arguments = raw_args
    else:
        arguments = raw_args

    return ToolCall(
        call_id=str(item.get("call_id", "")),
        name=str(item.get("name", "")),
        arguments=arguments,
    )


def _find_by_id(calls: list[ToolCall], call_id: str) -> ToolCall | None:
    for call in calls:
        if call.call_id == call_id:
            return call
    return None


def _extract_web_search_sources(item: dict[str, Any]) -> list[WebSearchSource]:
    action = item.get("action")
    if not isinstance(action, dict):
        return []

    raw_sources = action.get("sources")
    if not isinstance(raw_sources, list):
        return []

    sources: list[WebSearchSource] = []
    for source in raw_sources:
        if not isinstance(source, dict):
            continue
        url = str(source.get("url", "")).strip()
        if not url:
            continue
        title = _normalize_web_search_source_title(
            str(source.get("title", "")),
            url,
        )
        snippet = str(source.get("snippet", "")).strip()
        sources.append(
            WebSearchSource(
                title=title,
                url=url,
                snippet=snippet,
            )
        )
    return sources


def _display_web_search_result(
    display: DisplayProtocol,
    sources: list[WebSearchSource] | list[dict[str, str]],
    *,
    queries: list[str] | None = None,
) -> None:
    display.function_start(1)
    display.function_result(
        name="web_search",
        success=True,
        call_id="",
        arguments={
            "queries": [query for query in (queries or []) if query],
            "sources": (
                sources
                if sources and isinstance(sources[0], dict)
                else [
                    {
                        "title": source.title,
                        "url": source.url,
                        "snippet": source.snippet,
                    }
                    for source in sources
                ]
            ),
        },
    )
    display.tool_group_end()


def _extract_web_search_queries(item: dict[str, Any]) -> list[str]:
    action = item.get("action")
    if not isinstance(action, dict):
        return []

    raw_queries = action.get("queries")
    if isinstance(raw_queries, list):
        return [str(query).strip() for query in raw_queries if str(query).strip()]

    raw_query = action.get("query")
    if isinstance(raw_query, str) and raw_query.strip():
        return [raw_query.strip()]
    return []


def _annotation_web_search_source(
    annotation: dict[str, Any],
) -> WebSearchSource | None:
    url = str(annotation.get("url", "")).strip()
    if not url:
        return None
    return WebSearchSource(
        title=_normalize_web_search_source_title(str(annotation.get("title", "")), url),
        url=url,
    )


def _normalize_web_search_source_title(title: str, url: str) -> str:
    stripped = title.strip()
    if stripped and not _is_placeholder_web_search_title(stripped):
        return stripped
    parsed = urllib.parse.urlparse(url)
    host = parsed.netloc.strip().removeprefix("www.")
    return host or url


def _is_placeholder_web_search_title(title: str) -> bool:
    stripped = title.strip()
    if not stripped:
        return True
    return re.fullmatch(r"\[?\d+\]?", stripped) is not None


def _merge_web_search_sources(
    existing: list[WebSearchSource],
    incoming: list[WebSearchSource],
) -> list[WebSearchSource]:
    merged: list[WebSearchSource] = []
    by_url: dict[str, int] = {}

    for source in [*existing, *incoming]:
        url = source.url.strip()
        if not url:
            continue
        normalized = WebSearchSource(
            title=_normalize_web_search_source_title(source.title, url),
            url=url,
            snippet=source.snippet.strip(),
        )
        index = by_url.get(url)
        if index is None:
            by_url[url] = len(merged)
            merged.append(normalized)
            continue
        current = merged[index]
        current_title = _normalize_web_search_source_title(current.title, url)
        new_title = _normalize_web_search_source_title(normalized.title, url)
        merged[index] = WebSearchSource(
            title=current_title
            if not _is_placeholder_web_search_title(current_title)
            else new_title,
            url=url,
            snippet=current.snippet or normalized.snippet,
        )

    return merged


def _serialize_web_search_sources(
    sources: list[WebSearchSource],
) -> list[dict[str, str]]:
    return [
        {"title": source.title, "url": source.url, "snippet": source.snippet}
        for source in sources
    ]


def _deserialize_web_search_sources(raw_sources: Any) -> list[WebSearchSource]:
    if not isinstance(raw_sources, list):
        return []
    sources: list[WebSearchSource] = []
    for source in raw_sources:
        if not isinstance(source, dict):
            continue
        url = str(source.get("url", "")).strip()
        if not url:
            continue
        sources.append(
            WebSearchSource(
                title=_normalize_web_search_source_title(
                    str(source.get("title", "")),
                    url,
                ),
                url=url,
                snippet=str(source.get("snippet", "")).strip(),
            )
        )
    return sources


def _response_include(model: str) -> list[str]:
    include = ["web_search_call.action.sources"]
    if any(model.startswith(prefix) for prefix in _ENCRYPTED_REASONING_MODELS):
        include.append("reasoning.encrypted_content")
    return include


def _reasoning_request(model: str, effort: str) -> dict[str, Any]:
    if any(model.startswith(prefix) for prefix in _REASONING_EFFORT_MODELS):
        return {"effort": _EFFORT_MAP.get(effort, "high")}
    return {}


def _usage_value(usage_obj: Any, key: str) -> int:
    if not isinstance(usage_obj, dict):
        return 0
    return int(usage_obj.get(key, 0) or 0)


def _response_model_name(response_json: dict[str, Any]) -> str:
    model = response_json.get("model")
    return model if isinstance(model, str) else ""


def _read_error_body(exc: urllib.error.HTTPError) -> str:
    try:
        return exc.read().decode("utf-8", errors="replace")
    except Exception:
        return ""


def _normalize_http_error(
    exc: urllib.error.HTTPError,
    error_body: str,
) -> dict[str, Any]:
    payload = _parse_error_payload(error_body)
    error_type = _HTTP_ERROR_TYPES.get(exc.code)
    message = _HTTP_ERROR_MESSAGES.get(exc.code, f"HTTP {exc.code}")
    request_id = _request_id_from_headers(exc)

    if payload is not None:
        payload_request_id = payload.get("request_id")
        if isinstance(payload_request_id, str) and payload_request_id.strip():
            request_id = payload_request_id.strip()

        error_value = payload.get("error")
        if isinstance(error_value, dict):
            payload_type = error_value.get("type")
            if isinstance(payload_type, str) and payload_type.strip():
                error_type = payload_type.strip()
            payload_message = error_value.get("message")
            if isinstance(payload_message, str) and payload_message.strip():
                message = payload_message.strip()
        elif isinstance(error_value, str) and error_value.strip():
            message = error_value.strip()

    if error_type is None:
        if 400 <= exc.code < 500:
            error_type = "invalid_request_error"
        else:
            error_type = "api_error"

    return {
        "type": "error",
        "status": exc.code,
        "error": {
            "type": error_type,
            "message": message,
        },
        **({"request_id": request_id} if request_id else {}),
    }


def _parse_error_payload(error_body: str) -> dict[str, Any] | None:
    stripped = error_body.strip()
    if not stripped.startswith("{"):
        return None
    try:
        payload = json.loads(stripped)
    except json.JSONDecodeError:
        return None
    return payload if isinstance(payload, dict) else None


def _request_id_from_headers(exc: urllib.error.HTTPError) -> str | None:
    if not exc.headers:
        return None
    for header_name in ("request-id", "x-request-id"):
        request_id = exc.headers.get(header_name)
        if isinstance(request_id, str) and request_id.strip():
            return request_id.strip()
    return None


def _format_error_message(prefix: str, error_payload: dict[str, Any]) -> str:
    return f"{prefix}: {json.dumps(error_payload, sort_keys=True)}"


def _extract_retry_after(exc: urllib.error.HTTPError, attempt: int) -> float:
    try:
        retry_after = exc.headers.get("Retry-After") if exc.headers else None
        if retry_after:
            return max(0.1, min(float(retry_after), 60.0)) + 1.0
    except (TypeError, ValueError):
        pass
    return min(2.0 * (2**attempt), 30.0) + 1.0


def _waiting_message_for_input_items(input_items: list[dict[str, Any]]) -> str:
    has_user_message = any(
        isinstance(item, dict) and item.get("role") == "user" for item in input_items
    )
    if has_user_message:
        return "analyzing your request..."

    has_tool_output = any(
        isinstance(item, dict) and item.get("type") == "function_call_output"
        for item in input_items
    )
    if has_tool_output:
        return "integrating tool results..."

    return "processing..."


def _reasoning_body_text(reasoning_text: str, summary_text: str) -> str | None:
    if reasoning_text.strip():
        return reasoning_text
    body = _summary_body_text(summary_text)
    return body if body.strip() else None


def _summary_body_text(summary_text: str) -> str:
    lines = summary_text.splitlines()
    first_non_empty_index = next(
        (idx for idx, line in enumerate(lines) if line.strip()),
        None,
    )
    if first_non_empty_index is None:
        return ""
    remaining_non_empty = any(
        line.strip() for line in lines[first_non_empty_index + 1 :]
    )
    if not remaining_non_empty:
        return ""
    return "\n".join(lines[first_non_empty_index + 1 :]).lstrip("\r\n")
